#include "causal_path.h"


extern int print_pkt;
extern int print_id ;
extern int print_dns;
extern int test;
extern int dot_graph;
extern int remove_dup;
extern unsigned int numpkt;
extern int time_window;
extern int time_buffer_window;
extern bool print_execution_time;
extern int ambiguity_window;


Graphs::Graphs(){
	m_total_path=0;
	m_ambiguous_req=0;
	m_retransmission_req=0;
	m_tot_dht_req=0;
	m_tot_dns_req=0;
	m_tot_http_req=0;
}

TimeDiff::TimeDiff()
{
	m_tm[0].tv_sec = 0;
	m_tm[1].tv_sec = 0;
	m_tm[2].tv_sec = 0;
	m_tm[3].tv_sec = 0;

	m_tm[0].tv_usec = 0;
	m_tm[1].tv_usec = 0;
	m_tm[2].tv_usec = 0;
	m_tm[3].tv_usec = 0;
	
	m_flowCount = 0;
	m_type = DHT;

}



double TimeDiff::getOffsetTimeDiff(){
	long long int g = ((m_tm[2].tv_sec*1000000+m_tm[2].tv_usec) - (m_tm[1].tv_sec*1000000 + m_tm[1].tv_usec )) ;
	double offDouble = (double) g / 1000000.0 ; 
	return offDouble;

} 

double TimeDiff::getResponseTime(bool http){
	long long int g =0;
	double offDouble = 0;
	
	if ( http && (int)m_tm[2].tv_sec == 0 )
		return 0.0;

	if ( !http && (int)m_tm[1].tv_sec == 0 )
		return 0.0;

	if( m_type == HTTP_REQ )
		g = ((m_tm[2].tv_sec*1000000+m_tm[2].tv_usec) - (m_tm[0].tv_sec*1000000 + m_tm[0].tv_usec )) ;

	else if( m_type == DNS_REQ )
		g = ((m_tm[1].tv_sec*1000000+m_tm[1].tv_usec) - (m_tm[0].tv_sec*1000000 + m_tm[0].tv_usec )) ;
	
	offDouble = (double) g / 1000000.0 ; 
	return offDouble; 
	
}

double TimeDiff::getProcessingTimeDiff(){
	long long int g =0;
	double offDouble = 0;
	if ( (int)m_tm[1].tv_sec == 0 )
		return 0.0;
	
	if( m_type == HTTP_REQ || m_type == DNS_REQ ){
		g = ((m_tm[1].tv_sec*1000000+m_tm[1].tv_usec) - (m_tm[0].tv_sec*1000000 + m_tm[0].tv_usec )) ;
		offDouble = (double) g / 1000000.0 ; 
		return offDouble;
	}

	if ( (int)m_tm[2].tv_sec == 0 )
		return 0.0;
	
	if( m_type == DHT )
		g = ((m_tm[2].tv_sec*1000000+m_tm[2].tv_usec) - (m_tm[1].tv_sec*1000000 + m_tm[1].tv_usec )) ;
	
	offDouble = (double) g / 1000000.0 ; 
	return offDouble;

} 


double TimeDiff::getOutgoingLatency(){
	if(m_type != DHT )
		return 0;

	if ( (int)m_tm[2].tv_sec == 0 )
		return 0.0;

	long long int g =0;
	g = ((m_tm[1].tv_sec*1000000+m_tm[1].tv_usec) - (m_tm[0].tv_sec*1000000 + m_tm[0].tv_usec )) ;
	
	double offDouble = (double) g / 1000000.0 ; 
	return offDouble;
	
}

double TimeDiff::getIncomingLatency(){
	if(m_type != DHT )
		return 0;

	if ( (int)m_tm[2].tv_sec == 0 || (int)m_tm[3].tv_sec == 0)
		return 0.0;

	long long int g =0;
	g = ((m_tm[3].tv_sec*1000000+m_tm[3].tv_usec) - (m_tm[2].tv_sec*1000000 + m_tm[2].tv_usec )) ;
	
	double offDouble = (double) g / 1000000.0 ; 
	return offDouble;
	
}


/*
bool Graphs::checkThresTime(timeval* t1, timeval* t2, int time_thresh) 
{
	if (int(t1->tv_sec - t2->tv_sec) > time_thresh )
		return true;
	if (int(t1->tv_sec - t2->tv_sec) < time_thresh )
		return false;

	if ( t1->tv_usec < t2->tv_usec )
		return false;

	return true;
}

*/

struct timeval Graphs::time_difference(struct timeval &first, struct timeval &second) 
{ 
    struct timeval result; 
 
    if(first.tv_sec > second.tv_sec ||  
        (first.tv_sec == second.tv_sec &&  
        first.tv_usec > second.tv_usec)) { 
        result.tv_usec = first.tv_usec - second.tv_usec; 
        result.tv_sec = first.tv_sec - second.tv_sec; 
    } else { 
        result.tv_usec = second.tv_usec - first.tv_usec; 
        result.tv_sec = second.tv_sec - first.tv_sec; 
    } 
     
    /* For valid inputs, usec cannot be less than -999,999. */ 
    if(result.tv_usec < 0) { 
        result.tv_sec--; 
        result.tv_usec += 1000000; 
    } 
 
    return result; 
} 


void Graphs::print_header(struct pcap_pkthdr& hdr,	struct miphdr* &iph)
{
	unsigned int sip = ntohl(iph->saddr);
	unsigned int dip = ntohl(iph->daddr);
	printf ("%d ",(int)hdr.ts.tv_sec);
	printf_ip(sip);
	printf(" ");
	printf_ip(dip);
	printf(" ");
	fflush(stdout);
}

void Graphs::makeConnId(std::string &connid,unsigned int &ip, unsigned short &port)
{	
	int len = snprintf(NULL, 0, "%d.%d.%d.%d:%d", ip & 0xff, (ip >> 8) & 0xff, (ip >> 16) & 0xff, (ip >> 24) & 0xff, port);
	char http[len+1];
	snprintf(http,len+1,"%d.%d.%d.%d:%d", ip & 0xff, (ip >> 8) & 0xff, (ip >> 16) & 0xff, (ip >> 24) & 0xff, port);
	connid.assign(http);
}

void  Graphs::hexdump(std::string &data, unsigned char *cstr, int &len)
{
	char hexstr[2*len+1];
	hexstr[2*len] = '\0';
	for(int i=0;i<len;i++){
		sprintf(hexstr+2*i,"%.2x",cstr[i]);
	}
	data.assign(hexstr);
		
}

void Graphs::parseHttpReq(std::map<std::string, std::string> &prm, std::string& hdr, std::string& path )
{

  int pos_uri = hdr.find(" ",5);
  if (pos_uri == -1) return ;
  std::string uri = hdr.substr(5,pos_uri-5);
  if (uri.size() > 0)
    prm.insert(std::map<std::string,std::string>::value_type("uri" , uri));
  int pos_crlf = hdr.find("\r\n");
  if (pos_crlf == -1) return ;
  std::string line = hdr.substr(0,pos_crlf+2);
  //cout << line;
  pos_crlf += 2;

  while(1)
    {
      unsigned int pos_nline = hdr.find("\r\n",pos_crlf);
      if ( pos_nline == 0 || pos_nline >= hdr.size()) break;
      line = hdr.substr(pos_crlf,pos_nline-pos_crlf);
      //cout << line << endl;
      pos_crlf = pos_nline + 2;

      unsigned int pos_colon = line.find(":");
      std::string nam = line.substr(0, pos_colon);
      std::string val = line.substr(pos_colon + 2 );
      //cout << nam << "=" << val << endl;
      prm.insert(std::map<std::string,std::string>::value_type(nam, val));
      
    }
  path += "http://";
  std::map<std::string,std::string>::iterator iter = prm.find("Host");
  std::string host ;
  if( iter != prm.end() ) 
     host = iter->second;
  else{
  	iter = prm.find("host");
	if(iter != prm.end())
		host = iter->second;
  }
  if(print_id)
  {
  	printf("%s ",host.c_str());
  }

  unsigned int pos_cdnprfx = host.find("nyud.net");
  if ( pos_cdnprfx != 0 )
    path += host.substr(0,pos_cdnprfx - 1);

  iter = prm.find("uri");
  path += "/";
  if( iter != prm.end() ) {
    path += iter->second;
  }

  //cout  << path << endl;
  return;
}

void Graphs::parseTCPPkt(unsigned char * pktdata, int datalen, struct pcap_pkthdr& hdr, struct miphdr* &iph, struct mtcphdr* &tcphdr)
{
	if ( datalen == 0 ) return ;

  	std::string strdata((char*)pktdata,(size_t)5);

	if ( !(strdata.substr(0,4)=="GET ") && !(strdata.substr(0,4)=="HTTP") )
		return;


	if (strdata.substr(0,4)=="HTTP"){
		unsigned short dp = ntohs(tcphdr->dest);
		std::string connid;
		makeConnId( connid, iph->daddr, dp);
		matchHTTPRes(connid,hdr,iph);
		return;  
	}

	strdata.assign((char*)pktdata,(size_t)datalen);

	size_t pos = strdata.find("\r\n\r\n");
	if( pos == std::string::npos ){ 
		//Packet can be truncated 
		size_t pos_host = strdata.find("host");
		if ( pos_host == std::string::npos ){
			size_t pos_Host = strdata.find("Host");
			if( pos_Host == std::string::npos )
				return;
		}
		pos = datalen - 2;
	}

	std::string httpheader = strdata.substr (0,pos+2);
	std::string path;
	std::map<std::string, std::string> params;

	parseHttpReq(params,httpheader, path );
	unsigned short dp = ntohs(tcphdr->dest);
	unsigned short sp = ntohs(tcphdr->source);
	std::string httpid;

	makeConnId( httpid, iph->saddr, sp);
	updatePathID(path,hdr,iph,HTTP_REQ,httpid,dp);

	if(print_id)
	{
		//make_ID(path);
		str id (strbuf ("%s", path.c_str()));
		keyID hkey = makeID(id);
		strbuf s;
		s << hkey;
		print_header(hdr,iph);
		warnx << id << " " << s << " ";
		fflush(stderr);
		printf("%s\n",httpid.c_str());
		fflush(stdout);
	}
	//cout << httpheader;
	return;	
}


void Graphs::parseUDPPkt(unsigned char * pktdata, int datalen, struct pcap_pkthdr& hdr, struct miphdr* &iph, struct mudphdr* &udphdr)
{
	if ( datalen == 0 ) return ;
	unsigned short sp = ntohs(udphdr->source);
	unsigned short dp = ntohs(udphdr->dest);

	if ((sp == CORAL_DNS_PORT || dp == CORAL_DNS_PORT) )
    {	
    	//printf("PACKET NUMBER IS %d\n",numpkt);
		fflush(stdout);
      parse_dns(pktdata,datalen,hdr,iph,udphdr);
	  return;
    }

	std::string hexDataStr;
	hexdump(hexDataStr,(unsigned char *)pktdata,datalen);
	//printf("caplen = %d, datalen = %d, hdrsize = %d \n", hdr.caplen, datalen, hdr.caplen-datalen);
	if (print_pkt)
	{
		print_header(hdr,iph);
		printf("%s\n",hexDataStr.c_str());
	}
	matchDHTpacket( hexDataStr,  hdr, iph);

	return;

}

void Graphs::parse_dns(unsigned char * pktdata, int datalen, struct pcap_pkthdr& hdr,struct miphdr* &iph,struct mudphdr* &udphdr)
{
  if ( datalen == 0 ) return ;
  dns_request_t m; 
  memset(&m, 0, sizeof(m));
  
  m.numread = datalen;
  if (  datalen > (int)sizeof(m.original_buf) )
    return;
  
  memcpy(m.original_buf,pktdata,datalen);

  /* check that the message is long enough */ 
  if( m.numread < (int)sizeof (m.message.header) )
    return;

  //printf("parsing starting \n");

  dns_decode_request( &m ); 
  std::string cname (m.cname);
  std::string urlname ("http://");

  int pos_cdnprfx = cname.find("nyud.net");
  if ( pos_cdnprfx != -1 ){
    urlname += cname.substr(0,pos_cdnprfx - 1);
	urlname += "/";
	if(print_dns)
	{
		print_header(hdr,iph);
    	printf("%s %s %d\n",urlname.c_str(),m.ip,m.ttl);
	}
	unsigned short sp = ntohs(udphdr->source);
	unsigned short dp = ntohs(udphdr->dest);
	unsigned int sip = iph->saddr;
	pathType type = DNS;
	std::string httpid;
	if ( dp == CORAL_DNS_PORT ){
		type = DNS_REQ;
	}
	else{
		type = DNS_RES;
		sip = iph->daddr;
		sp = ntohs(udphdr->dest);
	}
	
	makeConnId( httpid, sip, sp);
	updatePathID(urlname,hdr,iph,type,httpid,dp);
  }


}


/******************************************************************************
*This Function update the ID and path from DNS and HTTP request parsing
*
*****************************************************************************/

void Graphs::updatePathID(std::string &urlid, struct pcap_pkthdr& hdr,struct miphdr* &iph,pathType type,std::string &httpid,unsigned short &dp)
{

	bool id_seen = false;
	bool is_ambiguous = false;
	bool isRetransmission = false;
	unsigned int sip = ntohl(iph->saddr);
	unsigned int dip = ntohl(iph->daddr);
	//Check for the ID in the list of IDs seen in last time_window seconds
	for (std::list<CausalPath *>::iterator it=m_path_list.begin(); it!=m_path_list.end(); ++it){
		CausalPath *cpath = *it;
	  	if (checkThresTime(&hdr.ts, &cpath->m_curTime,time_window)){
			it=m_path_list.erase(it);
			continue;
	  	}
		// if Id is found in last time_window sec, update the record and put it in the front of list
		//if(cpath->m_urlID.compare(urlid) == 0){
		if( cpath->isMatchID(urlid,sip,dip,type,is_ambiguous,httpid,isRetransmission,hdr)){
			cpath->m_curTime = hdr.ts;
		  	cpath->addFlow(sip, dip, hdr.ts,httpid,type);
			cpath->m_dht_ids[httpid]+=1;
		  	it = m_path_list.erase(it);
			
		  	id_seen = true;
			
			/*
			if(is_ambiguous)
				m_ambiguous_req ++;
			
			if(isRetransmission)
				m_retransmission_req ++;
			*/
			
		  	m_path_list.push_front(cpath);
			break;
	  }   
	}

	
	// IF Id is not seen in the last 5 sec. create a new ID and add it to the front of list
	if(!id_seen &&( dp == CORAL_DNS_PORT || dp == CORAL_HTTP_PORT )){
	  CausalPath *path = new CausalPath(urlid,sip,dip,hdr.ts,type,httpid);
	  m_path_list.push_front(path);
	  path->m_dht_ids[httpid]+=1;
	  m_path_time_map[++m_total_path] = path;

	  for (std::list<Flow *>::iterator it=m_dhtpkt_list.begin();it!=m_dhtpkt_list.end(); ++it){
		Flow *flow = *it;
		if (checkThresTime(&hdr.ts, &flow->m_srttime,time_buffer_window)){
			delete flow;
			it=m_dhtpkt_list.erase(it);
			continue;
	  	}
		break;
	  }
	  
	  for (std::list<Flow *>::reverse_iterator it=m_dhtpkt_list.rbegin();it!=m_dhtpkt_list.rend(); ++it){
		Flow *flow = *it;
		if( ! path->isMatchPath(flow->m_id,flow->m_sip,flow->m_dip) )
			continue;

		path->m_curTime = hdr.ts;
		flow->m_id = path->addDHTId(flow->m_id);
		path->m_flows.push_front(flow);
		//it=m_dhtpkt_list.erase(it);
		m_dhtpkt_list.remove(flow);
		if( it == m_dhtpkt_list.rend() )  break;
	  }
	}

}

/***************************************************************
* match HTTP response message to a path
***************************************************************/

void Graphs::matchHTTPRes(std::string &connid, struct pcap_pkthdr& hdr,struct miphdr* &iph)
{
	unsigned int sip = ntohl(iph->saddr);
	unsigned int dip = ntohl(iph->daddr);
	//Check for the ID in the list of IDs seen in last 5 seconds
	for (std::list<CausalPath *>::iterator it=m_path_list.begin(); it!=m_path_list.end(); ++it){
		CausalPath *cpath = *it;
		if (checkThresTime(&hdr.ts, &cpath->m_curTime,time_window)){
			it=m_path_list.erase(it);
			continue;
	  	}
		hash_map<std::string,int>::iterator iter;
		iter = cpath->m_dht_ids.find(connid);
		if (iter == cpath->m_dht_ids.end())
			continue;

		cpath->m_curTime = hdr.ts;
		pathType type = HTTP_RES;
		cpath->addFlow(sip, dip, hdr.ts,connid,type);

		if ( cpath->m_clietIP == dip )
		{
			timeval rtt = time_difference(cpath->m_curTime , cpath->m_startTime);
			cpath->m_rtt = rtt.tv_sec * 1000000 + rtt.tv_usec;
			cpath->m_isMatchResponse = true;
		}
		if ( cpath->coralWebIp == sip )
			cpath->m_isMatchResponse = true;
		
		it = m_path_list.erase(it);
		m_path_list.push_front(cpath);
		break;
		
	}
}


void Graphs::matchDHTpacket(std::string &hexData, struct pcap_pkthdr& hdr,struct miphdr* &iph)
{
	unsigned int sip = ntohl(iph->saddr);
	unsigned int dip = ntohl(iph->daddr);
	bool isPktMatched = false;
	pathType type = DHT;
	//Check for the ID in the list of IDs seen in last 5 seconds
	for (std::list<CausalPath *>::iterator it=m_path_list.begin(); it!=m_path_list.end(); ++it){
		CausalPath *cpath = *it;
		if (checkThresTime(&hdr.ts, &cpath->m_curTime,time_window)){
			it=m_path_list.erase(it);
			continue;
	  	}
		
		if( ! cpath->isMatchPath(hexData,sip,dip) )
			continue;
	  	cpath->m_curTime = hdr.ts;
		std::string id = cpath->addDHTId(hexData);
		cpath->addFlow(sip, dip, hdr.ts,id,type);
		it = m_path_list.erase(it);
		m_path_list.push_front(cpath);
		isPktMatched = true;
		break;
	}
	
	// Maintaining a window of last 1 sec DHT packet seen to handle time synchronization
	if(!isPktMatched){
		struct Flow *flw = new Flow(sip,dip,hdr.ts,hexData,type);
		m_dhtpkt_list.push_back(flw);
	}
	
	for (std::list<Flow *>::iterator  iter=m_dhtpkt_list.begin(); iter !=m_dhtpkt_list.end(); ++iter){
		Flow *flow = *iter;
		if (checkThresTime(&hdr.ts, &flow->m_srttime,time_buffer_window)){
			delete flow;
			iter=m_dhtpkt_list.erase(iter);
			if( iter == m_dhtpkt_list.end() )  break;
			continue;
	  	}
		else 
			break;
	}
}


bool CausalPath::isMatchID(std::string &urlid,unsigned int &sip, unsigned int &dip, pathType &type,bool &is_ambiguous,std::string &httpid,bool &isRetransmission, struct pcap_pkthdr& hdr)
{
	hash_map<std::string,int>::iterator it;
	it = m_dht_ids.find(httpid);

	if(m_urlID.compare(urlid) == 0){
		if(type == DNS_RES && sip==coralDnsIp && dip == m_clietIP) // for coral dns response packet
			return true;
		
		if(type == DNS_REQ && dip==coralDnsIp && sip == m_clietIP)// same client dns request are grouped together
			return true;
		
		if(type == HTTP_REQ ){ // for http request

			if(coralWebIp == 0){ // Request coming for the first time 
				coralWebIp = dip;
				return true;
			}
			if(coralWebIp == dip ){
				if(!m_isMatchResponse){
					//is_ambiguous = true;
					if(m_clietIP == sip)
						{m_isRetransmission = true;isRetransmission=true;}
					else{
						if (!checkThresTime(&hdr.ts, &m_startTime,ambiguity_window)){
							m_isAmbigous = true;is_ambiguous=true;
						}
						
					}
				}
				return true;
			}
			if (coralWebIp==sip )
				return true;
		}
	}
	return false;
}

bool CausalPath::isMatchPath(std::string &hexData, unsigned int &sip, unsigned int &dip)
{
	if ( coralWebIp != 0 ){
		if (coralWebIp != sip && coralWebIp != dip)
			return false;
	}
	
	size_t pos = hexData.find(m_sha1ID);
	std::string pkt_id = hexData.substr(0,DHTIDSIZE);
	hash_map<std::string,int>::iterator it;
	it = m_dht_ids.find(pkt_id);
	
	if ( pos == std::string::npos && it == m_dht_ids.end() )
		return false;

	return true;
}

std::string CausalPath::addDHTId(std::string &hexData)
{
	std::string id = hexData.substr(0,DHTIDSIZE);
	m_dht_ids[id]+=1;
	return id;
}


bool CausalPath::addFlow(unsigned int &sip, unsigned int &dip, struct timeval &srttime, std::string &id, pathType &type)
{
	struct Flow *flw = new Flow(sip,dip,srttime,id,type);
	m_flows.push_back(flw);
	return true;
}




void Graphs::printGraph()
{

	for(path_time_it it =m_path_time_map.begin();it!=m_path_time_map.end();++it){
		CausalPath *cpath = (*it).second;
		cpath->printPath();	
		
		hash_map<std::string, TimeDiff *> fid_Time_hashmap;
		cpath->getProcessingTime(fid_Time_hashmap);
		updateProcessingTime(fid_Time_hashmap,cpath);

		if (cpath->m_isAmbigous){
			printf("This causal path is ambigous\n");
			m_ambiguous_req ++;
		}
		if (cpath->m_isRetransmission ){
			printf("This causal path is retransmission \n");
			m_retransmission_req ++;
		}
		
		/*
		if (cpath->m_isAmbigous){
			printf{"This causal path is ambiguos\n");
		}
		*/
		
		//if ( print_execution_time ){
			cpath->printProcessingTime(fid_Time_hashmap);
		//}
		
		printf("==========================\n\n");
	}
	
	printf("=======================================================\n");
	printf("Total unique request = %d\n",m_total_path);
	printf("Total DHT request = %d\n",m_tot_dht_req);
	printf("Total DNS request = %d\n",m_tot_dns_req);
	printf("Total HTTP request = %d\n",m_tot_http_req);
	printf("Ambigous request = %d (%f)\n",m_ambiguous_req,(double)m_ambiguous_req/(double)m_tot_http_req);
	printf("Retransmission request = %d (%f)\n",m_retransmission_req,(double)m_retransmission_req/(double)m_tot_http_req);
	printf("=======================================================\n\n");

	printProcessingTimeStats();
}


void Graphs::printPTime(ip_procTime_hashmap &ipPTime_map,char *type)
{
	ip_procTime_hashmap::iterator ipter;
	int tot_req = 0;
	for(ipter=ipPTime_map.begin();ipter!=ipPTime_map.end();++ipter){
		printf("%s\n-----------------------\n",ipter->first.c_str());
		std::list<double> *pTimeList = ipter->second;
		std::list<double>::iterator lit;
		pTimeList->sort();
		double sumTime = 0;
		int size = (int)pTimeList->size();
		double median = 0;
		int count = 0;
		for ( lit = pTimeList->begin(); lit!=pTimeList->end(); ++lit ){
			count ++;
			double pTime = *lit;
			sumTime += pTime;
			if(count == (size/2 +1))
				median = pTime;
			if ( print_execution_time )
				printf("%.6f%c",pTime, ((count) %10)? ' ' : '\n');
		}
	
		double avgTime = sumTime/(double)size;
		double sum2 = 0;
		for(lit=pTimeList->begin(); lit!=pTimeList->end(); ++lit ){
			double pTime = *lit;
			sum2 += pow( pTime - avgTime, 2.0 );			
		}
		
		double s_dev = sqrt( sum2 / ( size ) );
		tot_req += size;
		//s_dev = sqrt( (( pow(sumTime,2.0)) -(( 1.0/size) * (pow(sumTime,2.0))))/ (size -1.0));
		//printf("\n# OF %s REQUEST PROCESSED %s = %d\n",type,ipter->first.c_str(),size);
		printf("\n\n[%s] %s :: SIZE = %d  MEAN = %.6f  MEDIAN = %.6f  MAX = %.6f  MIN = %.6f  SD = %.6f\n",type,ipter->first.c_str(),size,avgTime,median,pTimeList->back(),pTimeList->front(),s_dev);
		//printf("STANDARD DEVIATION = %.6f\n",s_dev);
		printf("\n------------------------\n");
	}
	printf("\n# OF %s REQUEST PROCESSED (All Coral Nodes)= %d\n",type,tot_req );


}

// This function print the Processing time statistics of all the coral nodes
void Graphs::printProcessingTimeStats()
{

	printf("STARTING DNS REQUEST PROCESSING TIME\n");
	printPTime(m_ipDnsProcessingTime_map,"DNS");
	printf("\n=======================================================\n");

	
	printf("STARTING HTTP REQUEST PROCESSING TIME\n");
	printPTime(m_ipHttpProcessingTime_map,"HTTP");
	printf("\n=======================================================\n");
	

	printf("STARTING DHT REQUEST PROCESSING TIME\n");
	printPTime(m_ipDhtProcessingTime_map,"DHT");
	printf("\n=======================================================\n");


	//Combine all the processing time into one m_ipDhtProcessingTime_map

	ip_procTime_hashmap::iterator ipter,it;

	for(ipter=m_ipDhtProcessingTime_map.begin();ipter!=m_ipDhtProcessingTime_map.end();++ipter){
		printf("%s\n-----------------------\n",ipter->first.c_str());
		std::list<double> *pTimeList = ipter->second;
		std::list<double>::iterator lit;

		it = m_ipHttpProcessingTime_map.find(ipter->first);
		if ( it != m_ipHttpProcessingTime_map.end() ){
			pTimeList->splice(pTimeList->end(),*(it->second));
		}

		it = m_ipDnsProcessingTime_map.find(ipter->first);
		if ( it != m_ipDnsProcessingTime_map.end() ){
			pTimeList->splice(pTimeList->end(),*(it->second));
		}
	}

	printf("STARTING COMBINED REQUEST PROCESSING TIME\n");
	printPTime(m_ipDhtProcessingTime_map,"COMBINE");
	printf("\n=======================================================\n");

	
}



CausalPath::CausalPath(std::string &urlid,unsigned int &sip, unsigned int &dip, struct timeval &stime, pathType &type,std::string &flowid)  
{
	init(urlid);
	//std::string s("");
	addFlow(sip, dip, stime,flowid, type);
	m_startTime = stime;
	m_curTime = stime;
	m_rtt = 0;
	m_isAmbigous = false;
	m_isRetransmission = false;
	m_isMatchResponse = false;
	if (type == HTTP_REQ){
		coralWebIp = dip;
		coralDnsIp = 0;
		m_clietIP = sip;
	}
	else if ( type == DNS_REQ ){
		coralDnsIp = dip;
		coralWebIp = 0;
		m_clietIP = sip;
	}

}


CausalPath::CausalPath(std::string &sid)  
{
	init(sid);
}

void CausalPath::init(std::string &sid)
{
	m_urlID = sid;
	str id (strbuf ("%s", m_urlID.c_str()));
	keyID hkey = makeID(id);
	strbuf s;
	s << hkey;
	str kid(s);
	m_sha1ID.assign(kid.cstr());
	//printf("%s - %s\n",m_urlID.c_str(),m_sha1ID.c_str());
}

void CausalPath::printPath()
{
	printf("PATHIDS :: %s - %s\n",m_urlID.c_str(),m_sha1ID.c_str());
	printf("FLOWS::\n");
	std::list<Flow *>::iterator it;
	
	if(dot_graph)
		printf( "digraph G {\n");
	int flcount = 1;
	
	hash_map<std::string,std::string> unq_flow;
	if (remove_dup){
		if(dot_graph)
			removeDupFlow(unq_flow,true);
		else
			removeDupFlow(unq_flow);
	}
	else{
		for ( it=m_flows.begin(); it!=m_flows.end(); ++it){
			Flow *flow = *it;
			flow->printFlow(flcount);
			flcount++;
		}
	}
	if(dot_graph)
		printf( "}\n");

	printf("Request RTT = %d\n",m_rtt);

}

// This function gets the Processing time of each node in the sequence
void CausalPath::getProcessingTime(hash_map<std::string, TimeDiff *> &fid_Time_hashmap)
{

	hash_map<std::string, TimeDiff *>::iterator fTimeIter;
	
	std::list<Flow *>::iterator it,it2;
	
	for ( it=m_flows.begin(),it2=m_flows.begin(); it!=m_flows.end(); ++it){
		if (++it2 == m_flows.end())
			it2 = it;
		Flow *flow = *it;
		Flow *nxtFlow = *(it2);
		std::string fid;
		flow->getUniqueFlowIDString(fid);

		

		fTimeIter = fid_Time_hashmap.find(flow->m_id);
		if(fTimeIter != fid_Time_hashmap.end()){
			TimeDiff *tdiff;
			tdiff = fid_Time_hashmap[flow->m_id] ;
			if(tdiff->m_flowCount > 3)
				continue;
			tdiff->m_tm[tdiff->m_flowCount ++] = flow->m_srttime;
			
		}
		else{
			TimeDiff *tdiff = new TimeDiff();
			fid_Time_hashmap[flow->m_id] = tdiff ;
			tdiff->m_tm[tdiff->m_flowCount ++] = flow->m_srttime;
			tdiff->m_type = flow->m_type;
			// For incoming http request, store the processing time of webserver
			// processing time = 1st DHT message - http request
			if(flow->m_type == HTTP_REQ && coralWebIp==flow->m_dip){ 
				if( coralWebIp==nxtFlow->m_sip )
					tdiff->m_tm[tdiff->m_flowCount++] = nxtFlow->m_srttime;
			}
		}
		
	}

}


// Update the processing time of current causal path Globally 
void Graphs::updateProcessingTime(hash_map<std::string, TimeDiff *> &fid_Time_hashmap,CausalPath* &cpath)
{
	hash_map<std::string, TimeDiff *> fid_Time_hashmap_tmp (fid_Time_hashmap);
	hash_map<std::string, TimeDiff *>::iterator fTimeIter;
	std::list<Flow *>::iterator it,it2;

	for ( it=cpath->m_flows.begin(),it2=cpath->m_flows.begin(); it!=cpath->m_flows.end(); ++it){
		Flow *flow = *it;
		fTimeIter = fid_Time_hashmap_tmp.find(flow->m_id);
		if(fTimeIter != fid_Time_hashmap_tmp.end()){
			TimeDiff *tdiff = fid_Time_hashmap_tmp[flow->m_id];
			std::string ip;
			ip_string(ip,flow->m_dip);
			ip_procTime_hashmap::iterator ipter;
			std::list<double> *pTimeList;
			if(flow->m_type == HTTP_REQ && cpath->coralWebIp == flow->m_dip){

				ipter = m_ipHttpProcessingTime_map.find(ip);

				if(ipter == m_ipHttpProcessingTime_map.end() ){
					pTimeList = new std::list<double>;
					m_ipHttpProcessingTime_map[ip]=pTimeList;
				}
				else{
					pTimeList = m_ipHttpProcessingTime_map[ip];
				}

				m_tot_http_req ++;
				if (tdiff->getProcessingTimeDiff() > 0 )
					pTimeList->push_back(tdiff->getProcessingTimeDiff());

			}

			if(flow->m_type == DHT ){

				ipter = m_ipDhtProcessingTime_map.find(ip);

				if(ipter == m_ipDhtProcessingTime_map.end() ){
					pTimeList = new std::list<double>;
					m_ipDhtProcessingTime_map[ip]=pTimeList;
				}
				else{
					pTimeList = m_ipDhtProcessingTime_map[ip];
				}
				m_tot_dht_req++;
				if (tdiff->getProcessingTimeDiff() > 0 )
					pTimeList->push_back(tdiff->getProcessingTimeDiff());

			}
			
			if(flow->m_type == DNS_REQ){

				ipter = m_ipDnsProcessingTime_map.find(ip);

				if(ipter == m_ipDnsProcessingTime_map.end() ){
					pTimeList = new std::list<double>;
					m_ipDnsProcessingTime_map[ip]=pTimeList;
				}
				else{
					pTimeList = m_ipDnsProcessingTime_map[ip];
				}
				m_tot_dns_req++;
				if (tdiff->getProcessingTimeDiff() > 0 )
					pTimeList->push_back(tdiff->getProcessingTimeDiff());

			}

			fid_Time_hashmap_tmp.erase(flow->m_id);
		}
	}
}

// Print the processing time for each causal path
void CausalPath::printProcessingTime(hash_map<std::string, TimeDiff *> &fid_Time_hashmap)
{

	hash_map<std::string, TimeDiff *>::iterator fTimeIter;
	
	std::list<Flow *>::iterator it,it2;
	double httpResTime =0,dnsResTime=0;
	printf("----------------------------------------------------------------------\n");
	printf("Request_type\t\tIPAddr\t\t\tProcessing_Time\tOutgoing_Latency\tIncoming_Latency\n");
	printf("-----------------------------------------------------------------------\n");
	for ( it=m_flows.begin(),it2=m_flows.begin(); it!=m_flows.end(); ++it){
		Flow *flow = *it;
		
		fTimeIter = fid_Time_hashmap.find(flow->m_id);
		if(fTimeIter != fid_Time_hashmap.end()){
			TimeDiff *tdiff = fid_Time_hashmap[flow->m_id];
			std::string ip;
			ip_string(ip,flow->m_dip);
			if (flow->m_type== HTTP_REQ && coralWebIp == flow->m_dip){
				printf("HTTP_REQUEST_CORAL [%s] = \t%.6f\n",ip.c_str(),tdiff->getProcessingTimeDiff());
				httpResTime = tdiff->getResponseTime(true);
			}
			if (flow->m_type== HTTP_REQ && coralWebIp == flow->m_sip){
				printf("HTTP_WEB_REQUEST   [%s] = \t%.6f\n",ip.c_str(),tdiff->getProcessingTimeDiff());
			}
			if (flow->m_type == DHT)
				printf("DHT_REQUEST_CORAL  [%s] = \t%.6f\t%.6f\t%.6f\n",ip.c_str(),tdiff->getProcessingTimeDiff(),tdiff->getOutgoingLatency(),tdiff->getIncomingLatency());

			if (flow->m_type== DNS_REQ && coralDnsIp== flow->m_dip){
				printf("DNS_REQUEST_CORAL  [%s] = \t%.6f\n",ip.c_str(),tdiff->getProcessingTimeDiff());
				dnsResTime = tdiff->getResponseTime(false);
			}

			delete tdiff;
			fid_Time_hashmap.erase(flow->m_id);
		}
	}
	if(httpResTime)
		printf("HTTP RESPONSE TIME  = %.6f\n", httpResTime);
	if(dnsResTime)
		printf("DNS RESPONSE TIME  = %.6f\n", dnsResTime);
}

//This function always prints the last of all the duplicate packets
void CausalPath::removeDupFlow(hash_map<std::string,std::string> &unq_flow,bool isdot)
{
	hash_map<unsigned int, std::string> ipType;
	hash_map<std::string,std::string>::iterator iter;
	std::list<std::string> flowids;
	int flnum = 1;
	for ( std::list<Flow *>::iterator it=m_flows.begin(); it!=m_flows.end(); ++it){
		Flow *flow = *it;
		
		std::string fid;
		std::string tmval;
		flow->getUniqueFlowIDString(fid);
		flow->getTimevalString(tmval);

		flowids.remove(fid);
		flowids.push_back(fid);
		
		iter = unq_flow.find(fid);
		if(iter == unq_flow.end()){
			std::string flwstr;
			flow->getPrintStr(flwstr,flnum++,isdot);
			unq_flow[fid] =  flwstr;
		}
		else{
			if ( isdot ){
				std::string flwstr;
				flow->getPrintStr(flwstr,flnum++,isdot);
				unq_flow[fid] =  flwstr;
			}
			else
				unq_flow[fid] =  tmval + unq_flow[fid];
		}

		if ( isdot )
			getIpTypeStr(ipType, flow);
	}

	for (std::list<std::string>::iterator it=flowids.begin();it!=flowids.end();++it){
		iter = unq_flow.find(*it);
		if(iter != unq_flow.end()){
			printf("%s\n",unq_flow[*it].c_str());
			unq_flow.erase(*it);
		}
	}

	if ( isdot ){
		std::string ip;
		for (hash_map<unsigned int,std::string>::iterator it=ipType.begin();it!=ipType.end();++it){
			
			ip_string(ip,it->first);
			printf("\t\"%s\"  %s\n",ip.c_str(),it->second.c_str());
		}
	}
	
}


void CausalPath::getIpTypeStr(hash_map<unsigned int, std::string> &ipType,Flow* &flow)
{
	// The following code will assign a color and type to each IP.
	hash_map<unsigned int,std::string>::iterator it, it2;
	std::string iptp;
	
	it = ipType.find(flow->m_sip);
	it2 = ipType.find(flow->m_dip);
	if(it != ipType.end() && it2 !=ipType.end())
		return;

	
	switch(flow->m_type){
		case(HTTP_REQ):
			if(coralWebIp== flow->m_dip){
				ipType[flow->m_sip]=iptp.assign("[label = \"Client\", color=blue]");
				ipType[flow->m_dip]=iptp.assign("[label = \"Coral + Web\", color=green,shape=box]");
			}
			else{
				ipType[flow->m_sip]=iptp.assign("[label = \"Coral + Web\", color=green,shape=box]");
				ipType[flow->m_dip]=iptp.assign("[label = \"Web\", color=black]");
			}
			break;
		case(HTTP_RES):
			if(coralWebIp== flow->m_sip){
				ipType[flow->m_dip]=iptp.assign("[label = \"Client\", color=blue]");
				ipType[flow->m_sip]=iptp.assign("[label = \"Coral + Web\", color=green,shape=box]");
			}
			else{
				ipType[flow->m_dip]=iptp.assign("[label = \"Coral + Web\", color=green,shape=box]");
				ipType[flow->m_sip]=iptp.assign("[label = \"web\", color=black]");
			}

			break;
		case(DNS_REQ):
			if(coralDnsIp == flow->m_dip){
				ipType[flow->m_sip]=iptp.assign("[label = \"Client\", color=blue]");
				ipType[flow->m_dip]=iptp.assign("[label = \"Coral + Dns\", color=green,shape=box]");
			}
			break;
		case(DNS_RES):
			if(coralDnsIp== flow->m_sip){
				ipType[flow->m_dip]=iptp.assign("[label = \"Client\", color=blue]");
				ipType[flow->m_sip]=iptp.assign("[label = \"Coral + Dns\", color=green,shape=box]");
			}
			break;
		case(DNS):
			break;
		case(DHT):
			if(it2== ipType.end())
				ipType[flow->m_dip]=iptp.assign("[label = \"Coral Node\", color=green,shape=box]");
			if(it== ipType.end())
				ipType[flow->m_sip]=iptp.assign("[label = \"Coral Node\", color=green,shape=box]");
			break;
		case(HTTP):
			break;
	
	}

}

void Flow::getUniqueFlowIDString(std::string &fid)
{
	int len = snprintf(NULL, 0, "%d:%d:%s", m_sip,m_dip,m_id.c_str());
	char id[len+1];
	snprintf(id,len+1, "%d:%d:%s", m_sip,m_dip,m_id.c_str());
	fid.assign(id);
	return;
}

void Flow::getTimevalString(std::string &tmval)
{
	int len = snprintf(NULL, 0, "%d.%.6d:", (int)m_srttime.tv_sec,(int)m_srttime.tv_usec);
	char tm[len+1];
	snprintf(tm,len+1, "%d.%.6d ", (int)m_srttime.tv_sec,(int)m_srttime.tv_usec);
	tmval.assign(tm);
	return;

}


void Flow::getPrintStr(std::string &flwstr,int flnum, bool isdot)
{
	std::string ptype;
	std::string color;
	getType(ptype,color);
	std::string sip;
	ip_string(sip,m_sip);
	std::string dip;
	ip_string(dip,m_dip);
	
	if ( isdot ){
		if ( ptype == "DHT")
		{
			ptype.assign("D");
			//ptype += m_id;
		}
		int len = snprintf(NULL, 0, "\t\"%s\" -> \"%s\" [label = \"%d:%s\",color=%s]",sip.c_str(),dip.c_str(),flnum,ptype.c_str(),color.c_str());
		char id[len+1];
		snprintf(id,len+1, "\t\"%s\" -> \"%s\" [label = \"%d:%s\",color=%s]",sip.c_str(),dip.c_str(),flnum,ptype.c_str(),color.c_str());
		flwstr.assign(id);
		return;
	}
	
	int len = snprintf(NULL, 0, "%d.%06d %s %s %s %s", (int)m_srttime.tv_sec, (int)m_srttime.tv_usec,sip.c_str(),dip.c_str(),ptype.c_str(),m_id.c_str());
	char id[len+1];
	snprintf(id,len+1, "%d.%06d %s %s %s %s", (int)m_srttime.tv_sec, (int)m_srttime.tv_usec,sip.c_str(),dip.c_str(),ptype.c_str(),m_id.c_str());
	flwstr.assign(id);
	return;
}

void Flow::printFlow(int &flnum)
{

	std::string ptype;
	std::string color;
	getType(ptype,color);
	
	if(!dot_graph){
	printf("%d.%06d ", (int)m_srttime.tv_sec, (int)m_srttime.tv_usec);
	printf_ip(m_sip);
	printf(" ");
	printf_ip(m_dip);
	printf(" %s",ptype.c_str());
	printf(" %s",m_id.c_str());
	printf("\n");
	}
	if (dot_graph)
	{
		printf("\t\"");
		printf_ip(m_sip);
		printf("\" -> \"");
		printf_ip(m_dip);
		printf("\" [label = \"%d:",flnum);
		if ( ptype == "DHT")
		{
			ptype.assign("D");
			ptype += m_id;
		}
		printf("%s\"];\n",ptype.c_str());
		//printf("\t\"");
		//printf_ip(m_sip);
		//printf("\" [label=\"%s\"];\n",m_id.c_str());
		//printf("\";\n");
	}
}



void Flow::getType(std::string &ptype, std::string &color)
{
	switch(m_type)
		{	
			case(DNS):
				ptype.assign("DNS");
				color.assign("cyan");
				break;
			case(DHT):
				ptype.assign("DHT");
				color.assign("red");
				break;
			case(HTTP):
				ptype.assign("HTTP");
				color.assign("blue");
				break;
			case(HTTP_REQ):
				ptype.assign("HTTP_REQ");
				color.assign("blue");
				break;
			case(HTTP_RES):
				ptype.assign("HTTP_RES");
				color.assign("blue");
				break;
			case(DNS_REQ):
				ptype.assign("DNS_REQ");
				color.assign("cyan");
				break;
			case(DNS_RES):
				ptype.assign("DNS_RES");
				color.assign("cyan");
				break;
		}
	return;

}

//Flow::~Flow(){}
Graphs::~Graphs()
{
	for (std::list<Flow *>::iterator  iter=m_dhtpkt_list.begin(); iter !=m_dhtpkt_list.end(); ++iter){
		Flow *flow = *iter;
		delete flow;
		iter=m_dhtpkt_list.erase(iter);
		if(iter==m_dhtpkt_list.end()) break;
	}
	
	for (std::list<CausalPath *>::iterator it=m_path_list.begin(); it!=m_path_list.end(); ++it){
		it=m_path_list.erase(it);
		if(it==m_path_list.end()) break;
	}
	
	for(path_time_it it =m_path_time_map.begin();it!=m_path_time_map.end();++it){
		CausalPath *cpath = (*it).second;
		delete cpath;
		m_path_time_map.erase((*it).first);
		if(it==m_path_time_map.end()) break;
	}
}

CausalPath::~CausalPath()
{
	for (std::list<Flow *>::iterator  iter=m_flows.begin(); iter !=m_flows.end(); ++iter){
		Flow *flow = *iter;
		delete flow;
		iter=m_flows.erase(iter);
		if(iter==m_flows.end()) break;
	}

	for(hash_map<std::string,int>::iterator it =m_dht_ids.begin();it!=m_dht_ids.end();++it){
		m_dht_ids.erase((*it).first);
		if(it==m_dht_ids.end()) break;
	}
}



