/*
 * Coral Packet Parser
 * author : Anup Goyal
 *
 * goal: Parse the network traces of coral web , dns and dhst traffic
 */

#include "causal_path.h"

int print_pkt = 0;
int print_id = 0;
int print_dns = 0;
int test = 0;
int print_paths = 0;
int dot_graph = 0;
int remove_dup = 0;
int time_window = 3;
int ambiguity_window = 1;
int time_buffer_window = 1;
unsigned int numpkt = 0;
bool print_execution_time = false;

char pcap_errbuf[PCAP_ERRBUF_SIZE];


void usage(char *pname)
{
	printf("Usage: %s -frwxzh \n"
	       "Parse the netwrok traces of coral and link the messages together\n"
	       "\n"
	       "-h		list this help message\n"
	       "-f [filter]	set a TCPDUMP filter to the input trace\n"
	       "-r [file]	read from the file. Default read from standard input (-)\n"
	       "-w [wfile]	write to the file prefix. The file names will be [wfile]-[SIP].\n"
	       "-p              write pkt hexdump in stdout\n"
	       "-i              write the http address and it corrosponding sha1 hash in stdout\n"
	       "-d              Write DNS request in stdout\n"
	       "-g				Write message linking graph in stdout\n"
	       "-c				write path in dot graph format\n"
	       "-m				Remove Duplicate flows\n"
	       "-e				Print the processing time for each flow\n"
	       "		Default wfile=file\n", pname);

	exit(0);
}

int main(int argc, char *argv[])
{
	pcap_t *pkt_d;
	char *pkt_filename = "-";
	char *dump_filename = NULL;

	char *pcap_filter = 0;
	struct bpf_program bpf_prog;
	struct pcap_pkthdr hdr;
	const u_char *pkt;
	int hdr_size;

	struct miphdr *iph;
	struct mudphdr *uh;
	struct mtcphdr *th;
	struct micmphdr *ich;

   	Graphs paths;
	
	map < connid_t, std::list < conntime_t > >conndict;
	map < connid_t, std::list < conntime_t > >::iterator connpos;
	std::list < conntime_t >::iterator tvit;

	map < unsigned int, vector < connrec_t > >scandict;
	map < unsigned int, vector < connrec_t > >::iterator scpos;
	vector < connrec_t >::iterator crit;


	char ch;
	while ((ch = getopt(argc, argv, "f:r:w:T:gmctipedh")) != -1) {
		switch (ch) {
		case 'f':
			if (strcmp(optarg, "") == 0) {
				fprintf(stderr, "Invalid parameter followed by -%c\n", ch);
				exit(1);
			}
			pcap_filter = optarg;
			break;
		case 'r':
			if (strcmp(optarg, "") == 0) {
				fprintf(stderr, "Invalid parameter followed by -%c\n", ch);
				exit(1);
			}
			pkt_filename = optarg;
			break;
		case 'w':
			if (strcmp(optarg, "") == 0) {
				fprintf(stderr, "Invalid parameter followed by -%c\n", ch);
				exit(1);
			}
			dump_filename = optarg;
			break;
		case 'T':
			if (strcmp(optarg, "") == 0) {
				fprintf(stderr, "Invalid parameter followed by -%c\n", ch);
				exit(1);
			}
			time_window = (int) atol(optarg);
			break;
		case 'p':
		  print_pkt = 1;
		  break;
		case 'i':
		  print_id = 1;
		  break;
		case 'd':
		  print_dns = 1;
		  break;
		case 't':
		  test = 1;
		  break;
		case 'g':
			print_paths = 1;
			break;
		case 'c':
			dot_graph = 1;
			break;
		case 'm':
			remove_dup= 1;
			break;
		case 'e':
			print_execution_time=true;
			break;
		case 'h':
			usage(argv[0]);
			break;
		}
	}

	argc -= optind;
	argv += optind;

	if ( test )
	  {
	    testEncoding();
		printf("time_window = %d, buffer window = %d\n",time_window,time_buffer_window);
	    return 0 ;
	  }

	if (!dump_filename) {
		if (strcmp(pkt_filename, "-") != 0)
			dump_filename = pkt_filename;
		else
			dump_filename = "senders";
	}

	if (!(pkt_d = pcap_open_offline(pkt_filename, pcap_errbuf))) {
		fprintf(stderr, "can't open file %s\n", pkt_filename);
		return -1;
	}

	if ((hdr_size = pcap_hdr_size(pkt_d)) < 0) {
		pcap_close(pkt_d);
		return -1;
	}

	if (pcap_filter) {
		if (pcap_compile(pkt_d, &bpf_prog, pcap_filter, 1, 0xffff0000))
			pcap_perror(pkt_d, "pcap_compile");
		else if (pcap_setfilter(pkt_d, &bpf_prog))
			pcap_perror(pkt_d, "pcap_setfilter");
	}

	connid_t cid;
	struct timeval tv;

	unsigned int numudppkt =0;
	int datalen =0;
	unsigned char *pktdata;

	while ((pkt = pcap_next(pkt_d, &hdr))) {
		++numpkt;
		//if(numpkt ==5 ) break;
		//printf("%d :: ",numpkt);
		iph = (struct miphdr *) (pkt + hdr_size);
		cid.sip = ntohl(iph->saddr);
		cid.dip = ntohl(iph->daddr);
		tv = hdr.ts;

		if (iph->protocol == IPPROTO_UDP) {
		  numudppkt ++;
		  
		  //if (print_pkt){
		  	//print_header(hdr,iph);
		  //}
		  
		  uh = (struct mudphdr *) (pkt + hdr_size + sizeof(struct miphdr));
		  unsigned int ulen = ntohs(uh->len);
		  if (ulen < 8)
		  	continue;
		  if (ulen > hdr.len)
		  	continue;
		  
		  cid.sp = ntohs(uh->source);
		  cid.dp = ntohs(uh->dest);
		  
		  datalen = hdr.caplen - ( hdr_size + sizeof(struct miphdr) + sizeof(struct mudphdr) );
		  //datalen = (int)ntohs(iph->tot_len) - (sizeof(struct miphdr) + sizeof(struct mudphdr) );
		  pktdata = (unsigned char *)(pkt + hdr_size + sizeof(struct miphdr) + sizeof(struct mudphdr));

			/*
		  if ((cid.sp == 53 || cid.dp == 53) )
		    {
		      paths.parse_dns(pktdata,datalen,hdr,iph);
			  //parse_dns(pktdata,datalen,hdr,iph);
		    }
		  else
		  */
		  	paths.parseUDPPkt(pktdata,datalen,hdr,iph,uh);
		    //extractData(pktdata,datalen,hdr,iph,IPPROTO_UDP);

		  
		  
		} else if (iph->protocol == IPPROTO_TCP) {
		  
		  th = (struct mtcphdr *) (pkt + hdr_size + sizeof(struct miphdr));
		  cid.sp = ntohs(th->source);
		  cid.dp = ntohs(th->dest);
		  int tcphdrlen = th->doff * 4;
		  datalen = hdr.caplen - ( hdr_size + sizeof(struct miphdr) + tcphdrlen );
		  //datalen = (int)ntohs(iph->tot_len) - (sizeof(struct miphdr) + tcphdrlen );
		  //printf ("%d\n",sizeof(struct mtcphdr)  );
		  pktdata = (unsigned char *)(pkt + hdr_size + sizeof(struct miphdr) + tcphdrlen);
		  //extractData(pktdata,datalen,hdr,iph,IPPROTO_TCP);
		  paths.parseTCPPkt(pktdata,datalen,hdr,iph,th);
		  
		} else if (iph->protocol == IPPROTO_ICMP) {
		  
		  ich = (struct micmphdr *) (pkt + hdr_size + sizeof(struct miphdr));
		  cid.sp = 0;
		  cid.dp = 0;

		  if (ich->type != ICMP_ECHO)
		    continue;
		  
		} else {
		  continue;
		}

	}
	
	if ( print_paths )
		paths.printGraph();
	/*

	dump_d=NULL;
	pos=dumpdict.find(string(dport));
	if (pos!=dumpdict.end()) {
	dump_d=pos->second;
	}
	
	if (!dump_d) {
	make_dump_filename(dump_fname,sizeof(dump_fname),dump_filename,dport);
	if (! (dump_d=pcap_dump_open(pkt_d,dump_fname))) {
	fprintf(stderr,"Cannot open file %s\n",dump_fname);
	pcap_close(pkt_d);
	exit(1);
	}
	dumpdict[string(dport)]=dump_d;
	}
	pcap_dump((u_char *) dump_d, &hdr, pkt);
	*/
	/* fprintf(stderr, "dump packet: %d.%06d (%d)\n", hdr.ts.tv_sec, hdr.ts.tv_usec, hdr.caplen); */
	
	/* Unimplemented yet! */
	//              if (post_process)
	//                      postprocess_dump_file(post_process,
	//                                            dump_filename2);
	pcap_close(pkt_d);
}

