#ifndef __HADOOP_HEADER__
#define __HADOOP_HEADER__

////////////////////////////// Rake needed functions ///////////////////////////
#include <vector>
#include <string>
using namespace std;

typedef vector<char> CharVect;

class JStream
{
public:
	JStream():data_(NULL),len_(0) {}
	JStream(char *data, int len):data_(data),len_(len) {}
	
	char *data_;
	int len_;

public:
	bool skip(int nbytes);
	bool skip_hrpc_header();
	bool readBytes(char *value, unsigned int len);
	bool readBool(bool &value);
	bool readShort(short & value);
	bool readUShort(unsigned short & value);
	bool readInt(int &value);
	bool readVInt(int & value);
	bool readLong(long long &value);
	bool readFloat(float & value);
	bool readString(string &value);
	bool readClassname(string &value);
	bool readJavaLangString(string &value);
		
};

//	Generate the Link_ID for this type of message
//	
//	
extern "C" CharVect ipc_call_id(char *pkt, int pkt_len);
extern "C" CharVect ipc_return_id(char *pkt, int pkt_len);
extern "C" CharVect sock_id(char *pkt, int pkt_len);
extern "C" vector<CharVect> sock_reverse_id(char *pkt, int pkt_len);

// DFS
string read_ipc_func_name(char *pkt, int iplen, JStream &stream);
CharVect get_filename_from_put_ipc_calls(char *pkt, int iplen, char *target_func_name);

extern "C" CharVect ipc_getFileInfo(char *pkt, int iplen);
extern "C" CharVect DFSFileInfo_ID(char *pkt, int iplen);
extern "C" CharVect ipc_getBlockLocations(char *pkt, int iplen);
extern "C" vector<CharVect> IPC_LocatedBlocks(char *pkt, int iplen);
extern "C" CharVect DFSClient_BlockReader(char *pkt, int iplen);
extern "C" bool Is_ReadBlock(char *pkt, int iplen);
extern "C" CharVect DFSClient_BlockReader_Done(char *pkt, int iplen);

extern "C" CharVect ipc_create(char *pkt, int iplen);
extern "C" CharVect ipc_addblock(char *pkt, int iplen);
extern "C" CharVect ipc_complete(char *pkt, int iplen);
extern "C" CharVect ipc_setreplication(char *pkt, int iplen);
extern "C" CharVect ipc_setpermission(char *pkt, int iplen);

extern "C" CharVect ipc_getSystemDir(char *pkt, int iplen);

// Job
extern "C" vector<CharVect> new_job_id(char *pkt, int iplen);

extern "C" CharVect get_jobid_from_mkdirs_ipc(char *pkt, int iplen);

CharVect get_jobid_from_upload_file_ipc(char *payload, int plen, const char *fname);
CharVect get_jobid_from_create_addblock_complete_ipc(char *pkt, int iplen, char* funcname, char*  filename);

extern "C" CharVect get_jobid_from_create_job_jar_ipc(char *pkt, int iplen);
extern "C" CharVect get_jobid_from_addblock_job_jar_ipc(char *pkt, int iplen);
extern "C" CharVect get_jobid_from_complete_job_jar_ipc(char *pkt, int iplen);
extern "C" CharVect get_jobid_from_setreplication_job_jar_ipc(char *pkt, int iplen);
extern "C" CharVect get_jobid_from_setpermission_job_jar_ipc(char *pkt, int iplen);

extern "C" CharVect get_jobid_from_create_job_xml_ipc(char *pkt, int iplen);
extern "C" CharVect get_jobid_from_addblock_job_xml_ipc(char *pkt, int iplen);
extern "C" CharVect get_jobid_from_complete_job_xml_ipc(char *pkt, int iplen);

extern "C" CharVect get_jobid_from_create_job_split_ipc(char *pkt, int iplen);
extern "C" CharVect get_jobid_from_addblock_job_split_ipc(char *pkt, int iplen);
extern "C" CharVect get_jobid_from_complete_job_split_ipc(char *pkt, int iplen);

extern "C" CharVect submitJob_ipc(char *pkt, int pkt_len);

extern "C" CharVect HeartBeatResponse_Launch_Task(char *pkt, int pkt_len);
extern "C" vector<CharVect> HeartBeatResponse_Launch_Task_TaskID(char *pkt, int pkt_len);

extern "C" CharVect heartbeat_status_succeeded(char *pkt, int pkt_len);
extern "C" CharVect heartbeat_status_running(char *pkt, int pkt_len);
extern "C" vector<CharVect> heartbeat_status_running_follow_ids(char *pkt, int pkt_len);

class ID
{
public:
	ID():id_(0) {}
	int id_;

	bool Read(JStream &stream);
	CharVect getID();
};

class JobID: public ID
{
public:
	string jtId_;
	bool isMap_;
	
	bool Read(JStream &stream);
	CharVect getJobID();
};

class Counters
{
public:
	int numClasses_;

	bool skip(JStream &stream);
};

class TaskStatus
{
public:
	ID taskAttemptID_;
	ID taskID_;
	JobID jobID_;

	float progress_;
	string runState_;
	string diagnosticInfo_;
	string stateString_;
	string phase_;
	long long startTime_;
	long long finishTime_;

	bool includeCounters_;
	
	bool Read(JStream &stream);
	CharVect getTaskID();
};

class Task
{
public:

	string jobFile_;
	ID taskAttemptID_;
	ID taskID_;
	JobID jobID_;
	int partition_;
	string outPath_;
	float progress_; 

	TaskStatus taskStatus_;
	
	bool Read(JStream &stream);
};

class TaskTrackerStatus
{
public:
	string trackerName_;
	string host_;
	int httpPort_;
	int failures_;
	vector<TaskStatus> taskReports_;

	//volatile long lastSeen;
	int maxMapTasks_;
	int maxReduceTasks_;

	bool Read(JStream &stream);
};

class InterTrackerProtocol_Heartbeat
{
public:
	TaskTrackerStatus status_;
	bool initialContact_;
	bool acceptNewTasks_;
	short responseId_;

	bool Read(JStream &stream);
};

class Block
{
public:
	long long blkid_;
	long long len_;
	long long generationStamp_;

	bool Read(JStream &stream);
};

class DatanodeID
{
public:
	string name_;
	string storageID_;
	int infoPort_;
	int ipcPort_;

	bool Read(JStream &stream);	
};

class DatanodeInfo: public DatanodeID
{
public:
	int ipcPort_;
	long long capacity_;
	long long dfsUsed_;
	long long remaining_;
	long long lastUpdate_;

	int xceiverCount_;
	string location_;
	string hostName_;
	string adminStates_;

	bool Read(JStream &stream);

};

class LocatedBlock
{
public:
	bool corrupt_;
	long long offset_;
	Block b_;
	vector<DatanodeInfo> locs_;

	bool Read(JStream &stream);
};

class LocatedBlocks
{
public:
	long long fileLength_;
	bool underConstruction_;
	vector<LocatedBlock> blocks_;
	
	bool Read(JStream &stream);
};

//extern "C" vector<CharVect> Follow_IDs(char *pkt, int pkt_len);

class DFSFileInfo
{
public:
	string path_;
	long long length_;
	bool isdir_;
	short block_replication_;
	long long blocksize_;
	long long modification_time_;
	short permission_;
	string owner_;
	string group_;

	bool Read(JStream &stream);
	void Print();
};

#endif

