#ifndef __HANDLE_PCAP_H__
#define __HANDLE_PCAP_H__

#include <sys/types.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <pcap.h>
#include <iostream>
#include <sstream>
#include <string>
#include <map>
#include <list>
#include <vector>
#include <algorithm>
#include <ext/hash_map>
#include <ext/hash_set>

using namespace std;
using namespace __gnu_cxx;

class HandlePcap {
public:
	HandlePcap();
	~HandlePcap();

	int pcap_hdr_size(pcap_t* pkt_d);
	int pcap_hdr_size();
	
	int status() { return status_; }
	bool open_pcap_file(char* &pkt_filename, char* &pcap_filter);
	bool open_pcap_file() {	return open_pcap_file(pcap_fname_, pcap_filter_); }
	bool close_pcap_file();

	
	const u_char* next_pkt(struct pcap_pkthdr *hdr);

	static inline void print_ip(unsigned int ip)
	{
		//ip = htonl(ip);
		printf("%d.%d.%d.%d", ip & 0xff, (ip >> 8) & 0xff, (ip >> 16) & 0xff, (ip >> 24) & 0xff);
	}

	static inline void print_ip(struct in_addr in)
	{
		unsigned int ip = in.s_addr; //htonl(in.s_addr);
		print_ip(ip);
	}

	// WAP5 related functions
	char *& pcap_fname() { return pcap_fname_; }
	char *& pcap_filter() { return pcap_filter_; }	
	in_addr_t  determine_local_address();
	in_addr_t  local_address() { return local_address_; };
	double avg_casual_delay();
	double lamda() { return lamda_; }
protected:
	pcap_t* pcap_;		// the handler

	char pcap_errbuf_[PCAP_ERRBUF_SIZE];

	int status_;
	int hdr_size_;

	char *pcap_fname_;
	char *pcap_filter_;
	
	// WAP5 related variables
	int delay_sample_num_;
	double avg_delay_;
	double lamda_; 		// lamda_ = 1/avg_delay_ 

	in_addr_t local_address_;

	double spontaneous_threshold_;
};

#endif

