/*
 * Coral Packet Parser
 * author : Anup Goyal
 *
 * goal: Parse the network traces of coral web , dns and dhst traffic
 */

#include "pkt_sync.h"

int print_pkt = 0;
int print_id = 0;
int print_dns = 0;
int test = 0;
int print_paths = 0;
int dot_graph = 0;
int remove_dup = 0;
bool verbose = false;
char pcap_errbuf[PCAP_ERRBUF_SIZE];
#define HOSTNAME 165.124.182.78
#define DHTIDSIZE 8  

int hostIP = -1518553526;
int num_pkt = 0;
int scrhostIP = -1518553526;
char *second_filename = "-";
char *pkt_filename = "-";
char *dump_filename = NULL;
char *srchost = "165.124.182.74";
char *dsthost = "-";



namespace __gnu_cxx                                                                                 
{                                                                                             
	template<> struct hash< std::string >                                                       
	{                                                                                           
		size_t operator()( const std::string& x ) const                                           
		{                                                                                         
			return hash< const char* >()( x.c_str() );                                              
		}                                                                                         
	};

}

class TimeDiff{
	public:
		timeval r;
		timeval s;
	public:
		TimeDiff();
		long long int getOffsetTimeDiff();
};

TimeDiff::TimeDiff()
{
	s.tv_sec = 0;
	r.tv_sec = 0;

	s.tv_usec = 0;
	r.tv_usec = 0;

}



long long int TimeDiff::getOffsetTimeDiff(){
	long long int g = ((r.tv_sec*1000000+r.tv_usec) - (s.tv_sec*1000000 + s.tv_usec )) ;
	return g;
} 



class TimeSyn{
	public:
		timeval s1;
		timeval r1;
		timeval s2;
		timeval r2;
		unsigned short check;
		std::string id;
	public:
		TimeSyn();
		long long int getOffsetTimeDiff();
};

TimeSyn::TimeSyn()
{
	s1.tv_sec = 0;
	r1.tv_sec = 0;
	s2.tv_sec = 0;
	r2.tv_sec = 0;

	s1.tv_usec = 0;
	r1.tv_usec = 0;
	s2.tv_usec = 0;
	r2.tv_usec = 0;
}

long long int TimeSyn::getOffsetTimeDiff(){
long long int g = ((r1.tv_sec*1000000+r1.tv_usec) - (s1.tv_sec*1000000 + s1.tv_usec )) - ((r2.tv_sec*1000000+r2.tv_usec) - (s2.tv_sec*1000000 + s2.tv_usec ));
g = g/2;
return g;
} 


/** BEGIN FIX **/
namespace __gnu_cxx 																				
{
	template<> struct hash<const TimeSyn * >                                                       
	{                                                                                           
		size_t operator()( const TimeSyn*  x ) const                                           
		{                                                                                         
			return hash< long >()( (long)x );                                              
		}                                                                                         
	};

	template<> struct hash<const TimeDiff* >                                                       
	{                                                                                           
		size_t operator()( const TimeDiff*  x ) const                                           
		{                                                                                         
			return hash< long >()( (long)x );                                              
		}                                                                                         
	};
}


hash_map<unsigned short, TimeSyn *> check_time_hashmap;
hash_map<std::string, TimeSyn *> id_time_hashmap;
hash_map<unsigned short, std::string> check_id_hashmap;


hash_map<std::string, TimeDiff *> check_inTime_hashmap;
hash_map<std::string, TimeDiff *> check_outTime_hashmap;

int pcap_hdr_size(pcap_t * pkt_d)
{
	int dl = pcap_datalink(pkt_d);
	int hdr_size = -1;

	switch (dl) {
	case DLT_NULL:
		hdr_size = 4;
		break;

	case DLT_EN10MB:
		hdr_size = 14;
		break;

	case DLT_FDDI:
		hdr_size = 13 + 8;	// fddi_header + LLC
		break;

	case DLT_RAW:
		hdr_size = 0;
		break;

	default:
		fprintf(stderr, "unknown data link type 0x%x", dl);
		break;
	}
	return hdr_size; 
}

struct timeval time_difference(struct timeval &first, struct timeval &second) 
{ 
    struct timeval result; 
 
    if(first.tv_sec > second.tv_sec ||  
        (first.tv_sec == second.tv_sec &&  
        first.tv_usec > second.tv_usec)) { 
        result.tv_usec = first.tv_usec - second.tv_usec; 
        result.tv_sec = first.tv_sec - second.tv_sec; 
    } else { 
        result.tv_usec = second.tv_usec - first.tv_usec; 
        result.tv_sec = second.tv_sec - first.tv_sec; 
    } 
     
    /* For valid inputs, usec cannot be less than -999,999. */ 
    if(result.tv_usec < 0) { 
        result.tv_sec--; 
        result.tv_usec += 1000000; 
    } 
 
    return result; 
} 


void hexdump(std::string &data, unsigned char *cstr, int &len)
{
	char hexstr[2*len+1];
	hexstr[2*len] = '\0';
	for(int i=0;i<len;i++){
		sprintf(hexstr+2*i,"%.2x",cstr[i]);
	}
	data.assign(hexstr);
		
}

void printf_ip(unsigned int ip)
{
	ip = htonl(ip);
	printf("%d.%d.%d.%d", ip & 0xff, (ip >> 8) & 0xff, (ip >> 16) & 0xff, (ip >> 24) & 0xff);
}


void print_header(struct pcap_pkthdr& hdr,	struct miphdr* &iph)
{
	unsigned int sip = ntohl(iph->saddr);
	unsigned int dip = ntohl(iph->daddr);
	printf ("%d ",(int)hdr.ts.tv_sec);
	printf_ip(sip);
	printf(" ");
	printf_ip(dip);
	printf(" ");
	fflush(stdout);
}


void printTimeDiff()
{
	hash_map<std::string, TimeDiff*>::iterator cit;
	long long int minOutTime = 100000000;
	long long int minInTime =  100000000;
	int numOutPair = 0;
	int numInPair = 0;
	for(cit=check_outTime_hashmap.begin();cit!=check_outTime_hashmap.end();++cit){
		TimeDiff *t1 = cit->second;
		if(t1->r.tv_sec == 0 || t1->s.tv_sec == 0 )
			continue;
		long long int outTime = t1->getOffsetTimeDiff();
		if ( verbose ) {
			printf("R1 = %d.%.6d\n",(int)t1->r.tv_sec,(int)t1->r.tv_usec);
			printf("S1 = %d.%.6d\n",(int)t1->s.tv_sec,(int)t1->s.tv_usec);
			cout<< ", TIME offset = " << outTime << endl;
			printf("================================\n");
		}
		if ( outTime < minOutTime )
			minOutTime = outTime;

		numOutPair++;
	}

	for(cit=check_inTime_hashmap.begin();cit!=check_inTime_hashmap.end();++cit){
		TimeDiff *t1 = cit->second;
		if(t1->r.tv_sec == 0 || t1->s.tv_sec == 0 )
			continue;
		long long int inTime = t1->getOffsetTimeDiff();
		if ( verbose ) {
			printf("R2 = %d.%.6d\n",(int)t1->r.tv_sec,(int)t1->r.tv_usec);
			printf("S2 = %d.%.6d\n",(int)t1->s.tv_sec,(int)t1->s.tv_usec);
			cout << ", TIME offset = " << inTime << endl;
			printf("================================\n");
		}
				
		if ( inTime < minInTime )
			minInTime = inTime;

		numInPair ++;
	}

	long long int offsetDiff = (minOutTime - minInTime)/2;

	cout << "# of outgoing going request (R1 - S1)= " << check_outTime_hashmap.size() << " ,Pairs = "<< numOutPair << endl;
	cout << "# of incoming going request (R2 - S2)= " << check_inTime_hashmap.size() << " ,Pairs = "<< numInPair << endl;
	cout << "Minumum of (R1 - S1) = " << minOutTime << endl;
	cout << "Minumum of (R2 - S2) = " << minInTime << endl;
	cout << "Time OFFset is ( (R1 - S1) - (R2 - S2) )/2 = " << offsetDiff << "   in (usec)" << endl;
	
	double offDouble = - (double) offsetDiff / 1000000.0 ; 

	printf("Filename = %s , fileIP = %s, Offset = %.6f \n", second_filename, dsthost, offDouble);
	
	printf("editcap -t%.6f %s %s.syn.pcap\n", offDouble, second_filename,second_filename);
	
	/*
	hash_map<std::string,TimeSyn *>::iterator it;
	long long int sumdiff = 0;
	int count = 0;
	for( it=id_time_hashmap.begin();it!=id_time_hashmap.end();++it){

		TimeSyn *t1 = it->second;

		if(t1->r1.tv_sec == 0 || t1->s1.tv_sec == 0 || t1->s2.tv_sec == 0 || t1->r2.tv_sec == 0  )
			continue;

		count ++;
		long long int tdiff = t1->getOffsetTimeDiff();
		sumdiff += tdiff;
		
		if ( verbose ) {
			printf("R1 = %d.%.6d\n",(int)t1->r1.tv_sec,(int)t1->r1.tv_usec);
			printf("S1 = %d.%.6d\n",(int)t1->s1.tv_sec,(int)t1->s1.tv_usec);
			printf("R2 = %d.%.6d\n",(int)t1->r2.tv_sec,(int)t1->r2.tv_usec);
			printf("S2 = %d.%.6d\n",(int)t1->s2.tv_sec,(int)t1->s2.tv_usec);
			cout<< "TIME offset = " << tdiff << endl;
			printf("================================\n");
		}

	}
	if ( count > 0)
		sumdiff = sumdiff/count;

	cout << "AVERAGE TIME OFFSET = " <<  sumdiff << "  (usec)" << endl;
	cout << "Number of offset analyzed = " << count << endl;
	*/
}

void parseUDPPkt(unsigned char * pktdata, int datalen, struct pcap_pkthdr& hdr, struct miphdr* &iph, struct mudphdr* &udphdr,bool second=false)
{
	if ( datalen == 0 ) return ;
	num_pkt++;
	unsigned int sip = iph->saddr;
	unsigned int dip = iph->daddr;
	//printf("%d, %d\n",sip,dip);
	std::string hexDataStr;
	hexdump(hexDataStr,(unsigned char *)pktdata,datalen);
	//printf("caplen = %d, datalen = %d, hdrsize = %d \n", hdr.caplen, datalen, hdr.caplen-datalen);
	if (print_pkt)
	{
		print_header(hdr,iph);
		printf("%s\n",hexDataStr.c_str());
	}

	u_int16_t check = udphdr->check;
	TimeDiff *t1;
	hash_map<std::string, TimeDiff*>::iterator cit;

	
	//if ( sip == (unsigned int)hostIP ){
	if ( sip == scrhostIP ){
		cit = check_outTime_hashmap.find(hexDataStr);
			
		if (  cit == check_outTime_hashmap.end() ){
			t1 = new TimeDiff();
			check_outTime_hashmap[hexDataStr]=t1;
		}
		else{
			t1 = check_outTime_hashmap[hexDataStr];
		}
		if ( !second ){
			t1->s = hdr.ts;
		}
		else
			t1->r = hdr.ts;
	}

	//if ( dip == (unsigned int)hostIP ){
	if ( dip == scrhostIP){
		cit = check_inTime_hashmap.find(hexDataStr);
		
		if (  cit == check_inTime_hashmap.end() ){
			t1 = new TimeDiff();
			check_inTime_hashmap[hexDataStr]=t1;
		}
		else{
			t1 = check_inTime_hashmap[hexDataStr];
		}
		
		if ( !second ){
			t1->r = hdr.ts;
		}
		else
			t1->s = hdr.ts;
	}

	


	/*
	std::string id = hexDataStr.substr(0,DHTIDSIZE);
	hash_map<std::string,TimeSyn *>::iterator it;
	it = id_time_hashmap.find(id);


	if(!second){
		TimeSyn *t1;
		
		if (  it == id_time_hashmap.end() ){
			t1 = new TimeSyn();
			t1->id = id;
			t1->check = check;
			check_time_hashmap[check]=t1;
			id_time_hashmap[id] = t1;


		}
		else{
			t1 = id_time_hashmap[id];
			check_time_hashmap[check] = t1;
		}

		
		if ( sip == (unsigned int)hostIP ){
			t1->s1 = hdr.ts;
		}
		if ( dip == (unsigned int)hostIP ){
			t1->r2 = hdr.ts;
		}
	}

	// Packets from second interface
	if(second){
		hash_map<unsigned short, TimeSyn *>::iterator cit;
		cit = check_time_hashmap.find(check);
		
		if (  it != id_time_hashmap.end() ){
			
			TimeSyn *t1 = id_time_hashmap[id];
			
			if ( sip == (unsigned int)hostIP ){
				t1->r1 = hdr.ts;
			}
			if ( dip == (unsigned int)hostIP ){
				t1->s2 = hdr.ts;
			}
		}
	}
	*/

	
	return;

}


void openPcapFile(char* &pkt_filename, char* &pcap_filter, int flNum){

	pcap_t *pkt_d;


	struct bpf_program bpf_prog;
	struct pcap_pkthdr hdr;
	const u_char *pkt;
	int hdr_size;

	struct miphdr *iph;
	struct mudphdr *uh;
	struct mtcphdr *th;
	struct micmphdr *ich;

	bool second = false;

	if (flNum > 1)
		second = true;
	
	if (!(pkt_d = pcap_open_offline(pkt_filename, pcap_errbuf))) {
		fprintf(stderr, "can't open file %s\n", pkt_filename);
		exit(-1);
	}

	if ((hdr_size = pcap_hdr_size(pkt_d)) < 0) {
		pcap_close(pkt_d);
		exit(-1);
	}

	if (pcap_filter) {
		if (pcap_compile(pkt_d, &bpf_prog, pcap_filter, 1, 0xffff0000))
			pcap_perror(pkt_d, "pcap_compile");
		else if (pcap_setfilter(pkt_d, &bpf_prog))
			pcap_perror(pkt_d, "pcap_setfilter");
	}

	connid_t cid;
	struct timeval tv;
	unsigned int numpkt = 0;
	unsigned int numudppkt =0;
	int datalen =0;
	unsigned char *pktdata;

	while ((pkt = pcap_next(pkt_d, &hdr))) {
		++numpkt;

		iph = (struct miphdr *) (pkt + hdr_size);
		cid.sip = ntohl(iph->saddr);
		cid.dip = ntohl(iph->daddr);
		tv = hdr.ts;

		if (iph->protocol == IPPROTO_UDP) {
		  numudppkt ++;
		  
		  uh = (struct mudphdr *) (pkt + hdr_size + sizeof(struct miphdr));
		  cid.sp = ntohs(uh->source);
		  cid.dp = ntohs(uh->dest);
		  int caplen = hdr.caplen;
		  if( hdr.caplen > 500 )
		  	caplen = 500;
		  //datalen = (int)ntohs(iph->tot_len) - (sizeof(struct miphdr) + sizeof(struct mudphdr) );
		  datalen = caplen - (hdr_size + sizeof(struct miphdr) + sizeof(struct mudphdr) );
		  pktdata = (unsigned char *)(pkt + hdr_size + sizeof(struct miphdr) + sizeof(struct mudphdr));

		  	parseUDPPkt(pktdata,datalen,hdr,iph,uh,second);
		  
		  
		} else if (iph->protocol == IPPROTO_TCP) {
		  
		  th = (struct mtcphdr *) (pkt + hdr_size + sizeof(struct miphdr));
		  cid.sp = ntohs(th->source);
		  cid.dp = ntohs(th->dest);
		  int tcphdrlen = th->doff * 4;
		  datalen = (int)ntohs(iph->tot_len) - (sizeof(struct miphdr) + tcphdrlen );
		  pktdata = (unsigned char *)(pkt + hdr_size + sizeof(struct miphdr) + tcphdrlen);
		} else if (iph->protocol == IPPROTO_ICMP) {
		  
		  ich = (struct micmphdr *) (pkt + hdr_size + sizeof(struct miphdr));
		  cid.sp = 0;
		  cid.dp = 0;

		  if (ich->type != ICMP_ECHO)
		    continue;
		  
		} else {
		  continue;
		}

	}
	
	/*

	dump_d=NULL;
	pos=dumpdict.find(string(dport));
	if (pos!=dumpdict.end()) {
	dump_d=pos->second;
	}
	
	if (!dump_d) {
	make_dump_filename(dump_fname,sizeof(dump_fname),dump_filename,dport);
	if (! (dump_d=pcap_dump_open(pkt_d,dump_fname))) {
	fprintf(stderr,"Cannot open file %s\n",dump_fname);
	pcap_close(pkt_d);
	exit(1);
	}
	dumpdict[string(dport)]=dump_d;
	}
	pcap_dump((u_char *) dump_d, &hdr, pkt);
	*/
	/* fprintf(stderr, "dump packet: %d.%06d (%d)\n", hdr.ts.tv_sec, hdr.ts.tv_usec, hdr.caplen); */
	
	/* Unimplemented yet! */
	//              if (post_process)
	//                      postprocess_dump_file(post_process,
	//                                            dump_filename2);
	pcap_close(pkt_d);


}

void usage(char *pname)
{
	printf("Usage: %s -frwxzh \n"
	       "Parse the netwrok traces of coral and link the messages together\n"
	       "\n"
	       "-h		list this help message\n"
	       "-f [filter]	set a TCPDUMP filter to the input trace\n"
	       "-r [file]	read from the file. Default read from standard input (-)\n"
	       "-w [wfile]	write to the file prefix. The file names will be [wfile]-[SIP].\n"
	       "-H [IP_ADDRESS] set the reference IP address\n"
	       "-p              write pkt hexdump in stdout\n"
	       "-i              source ip\n"
	       "-d              destination host\n"
	       "-g				Write message linking graph in stdout\n"
	       "-c				write path in dot graph format\n"
	       "-m				Remove Duplicate flows\n"
	       "-v 				verbose output\n"
	       "-s 				Second pcap file \n"
	       "		Default wfile=file\n", pname);

	exit(0);
}

int main(int argc, char *argv[])
{


	char *pcap_filter = 0;


	
	map < connid_t, std::list < conntime_t > >conndict;
	map < connid_t, std::list < conntime_t > >::iterator connpos;
	std::list < conntime_t >::iterator tvit;

	map < unsigned int, vector < connrec_t > >scandict;
	map < unsigned int, vector < connrec_t > >::iterator scpos;
	vector < connrec_t >::iterator crit;


	char ch;
	while ((ch = getopt(argc, argv, "f:r:w:s:i:d:gmctpvh")) != -1) {
		switch (ch) {
		case 'f':
			if (strcmp(optarg, "") == 0) {
				fprintf(stderr, "Invalid parameter followed by -%c\n", ch);
				exit(1);
			}
			pcap_filter = optarg;
			break;
		case 'r':
			if (strcmp(optarg, "") == 0) {
				fprintf(stderr, "Invalid parameter followed by -%c\n", ch);
				exit(1);
			}
			pkt_filename = optarg;
			break;
		case 'w':
			if (strcmp(optarg, "") == 0) {
				fprintf(stderr, "Invalid parameter followed by -%c\n", ch);
				exit(1);
			}
			dump_filename = optarg;
			break;
		case 's':
			if (strcmp(optarg, "") == 0) {
				fprintf(stderr, "Invalid parameter followed by -%c\n", ch);
				exit(1);
			}
			second_filename = optarg;
			break;
		case 'p':
		  print_pkt = 1;
		  break;
		case 'i':
			if (strcmp(optarg, "") == 0) {
				fprintf(stderr, "Invalid parameter followed by -%c\n", ch);
				exit(1);
			}
			srchost = optarg;
		  print_id = 1;
		  break;
		case 'd':
			if (strcmp(optarg, "") == 0) {
				fprintf(stderr, "Invalid parameter followed by -%c\n", ch);
				exit(1);
			}
			dsthost = optarg;
		  print_dns = 1;
		  break;
		case 't':
		  test = 1;
		  break;
		case 'g':
			print_paths = 1;
			break;
		case 'c':
			dot_graph = 1;
			break;
		case 'm':
			remove_dup= 1;
			break;
		case 'v':
			verbose= true;
			break;
		case 'h':
			usage(argv[0]);
			break;
		}
	}

	argc -= optind;
	argv += optind;

	if (!dump_filename) {
		if (strcmp(pkt_filename, "-") != 0)
			dump_filename = pkt_filename;
		else
			dump_filename = "senders";
	}

	if (strcmp(dsthost,"-")==0){
		fprintf(stderr,"please enter the destination host with -d option");
		exit(1);
	}
	scrhostIP = (int) inet_addr(srchost); 
	printf("dst IP = %s, src IP = %s (%d) \n", dsthost,srchost,scrhostIP);
	int len = snprintf(NULL, 0, "udp and ((src host %s or %s) and (dst host %s or %s))", srchost,dsthost,srchost,dsthost);
	char pcap_fil[len+1];
	snprintf(pcap_fil,len+1, "udp and ((src host %s or %s) and (dst host %s or %s))", srchost,dsthost,srchost,dsthost);
	printf("%s\n",pcap_fil);
	pcap_filter = pcap_fil;
	
	openPcapFile(pkt_filename, pcap_filter, 1);
	openPcapFile(second_filename, pcap_filter, 2);

	printTimeDiff();
	cout << "Total Number of packets are " << num_pkt << endl;
}


