/*
 * Rake
 * Author: 
 *	Yao Zhao
 * Goal:
 *	Implement Rake
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <math.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <dlfcn.h>
#include <regex.h>

#include "xmlParser.h"
#include "handle_pcap.h"
#include "rake.h"
#include "util.h"
#include "dns.h"

/////////////////////////////////////////////////////////

int Expression::read_int(const char *expr, int &value)
{
	int i = 0, j=0;
	cout <<"read_int: " << expr << endl;
	if( expr[0] == '-' ) {
		j++;
		expr++;
	}
	
	for(i=0; ; i++) {
		if(expr[i] == 0)
			break;
		
		if(expr[i] < '0' || expr[i] > '9')
			break;		
	}

	if(i==0)
		return -1;
	
	value = atoi( expr );
	
	return i+j;
}

int Expression::read_offset(const char *expr, int &value)
{
	int end = 0;
	if( *expr != '[' )
		return -1;

	expr++;

	for(int i=0; ; i++) {
		//cout << "i: " << i << endl;
		if(expr[i] == 0)
			return -1;

		if( expr[i] == ']' ) {
			if(i == 0)
				return -1;
			
			end = i;
			break;
		}
		else if(expr[i] < '0' || expr[i] > '9')
			return -1;		
	}

	value = atoi( expr );
	//cout << "value: " << value << ", end: " << end << endl;
	return end+2;
}

void Expression::Print(const ExprItem & item)
{
	if(item.type_ == EXPR_TYPE_OPERATION) {
		printf( "o: %d\n", item.data_.operation_);
	}
	else if(item.type_ == EXPR_TYPE_INT)
		printf( "i: %d\n", item.data_.ivalue_);
	else if(item.type_ == EXPR_TYPE_PACKET)
		printf("p: %d %d\n", item.data_.pvalue_.start_, item.data_.pvalue_.offset_);
}

void Expression::PrintItems()
{
	cout << endl << "Print expression" << endl;
	
	for(unsigned int i=0; i<items_.size(); i++) {
		Print(items_[i]);
	}
	cout << endl;
}

bool Expression::SetExpr(const char *expr)
{
	while(true) {
		int ret = ReadNextItem(expr);

		if(ret == 0)
			return true;

		if(ret <0)
			return false;
	}
}

int Expression::ReadNextItem(const char* &expr)
{
	ExprItem ei;
	int len = 0;
	
	switch(*expr) {
		case '\0':
			return 0;
			
		case ' ':
		case '\t':
			expr++;
			return 1;
		case '+':
			ei.data_.operation_ = OPERATION_PLUS;
			ei.type_ = EXPR_TYPE_OPERATION;
			expr++;
			break;
		case '-':
			ei.data_.operation_ = OPERATION_MINUS;
			ei.type_ = EXPR_TYPE_OPERATION;
			expr++;
			break;
		case '/':
			ei.type_ = EXPR_TYPE_OPERATION;
			ei.data_.operation_ = OPERATION_DIVIDE;
			expr++;
			break;
		case '*':
			ei.type_ = EXPR_TYPE_OPERATION;
			ei.data_.operation_ = OPERATION_MULTIPLY;
			expr++;
			break;
		case '|':
			ei.type_ = EXPR_TYPE_OPERATION;
			ei.data_.operation_ = OPERATION_OR;
			expr++;
			break;
		case '&':
			ei.type_ = EXPR_TYPE_OPERATION;
			ei.data_.operation_ = OPERATION_AND;
			expr++;
			break;

		case '=':
			ei.type_ = EXPR_TYPE_OPERATION;
			ei.data_.operation_ = OPERATION_EQUAL;
			expr++;
			break;

		case '<':
			ei.type_ = EXPR_TYPE_OPERATION;
			if( expr[1] != '=' ) {
				ei.data_.operation_ = OPERATION_LESS;			
				expr++;
			}
			else {
				ei.data_.operation_ = OPERATION_LESSEQUAL;
				expr+=2;
			}
			break;
		case '>':
			ei.type_ = EXPR_TYPE_OPERATION;
			if( expr[1] != '=' ) {
				ei.data_.operation_ = OPERATION_GREATER;			
				expr++;
			}
			else {
				ei.data_.operation_ = OPERATION_GREATEREQUAL;
				expr+=2;
			}
			break;

		case 'p':
			if(strncmp(expr, "payload", 7) == 0) {				
				ei.type_ = EXPR_TYPE_PACKET;
				ei.data_.pvalue_.start_ = 0;
								
				if( (len = read_offset( expr + 7, ei.data_.pvalue_.offset_ )) <= 0) {
					cerr << "Bad payload format: " << expr << endl;
					return -1;
				}

				expr += 7 + len;				
			}
			break;
			
		default:
			if( (len = read_int(expr, ei.data_.ivalue_)) > 0 ) {
				ei.type_ = EXPR_TYPE_INT;
				
				expr += len;
				cout << "int" << endl;
			}
			else { 
				cerr << "Bad int format: " << expr << endl;
					return -1;
			}
			break;
	}

	items_.push_back( ei );

	return 1;
}


//////////////////////// Rule /////////////////////////////

ClassifyRule::ClassifyRule():protocol_(-1),port_(0),sport_(0),dport_(0),start_pos_(REGEX_START_PAYLOAD)
{
	
}

bool ClassifyRule::classify(char *pkt, int pkt_len)
{
	struct ip *iph = (struct ip*) pkt;
	struct udphdr *udph;
	struct tcphdr *tcph;

	unsigned short sport = 0, dport = 0;
	//bool bMatch = true;
	
	char *transport_pkt = ((char *)iph) + iph->ip_hl*4;
	udph = (struct udphdr *) transport_pkt;
	tcph = (struct tcphdr *) transport_pkt;
	
	//printf("In classify. Protocol: %d. Rules number: %d\n", iph->ip_p, rules_.size());
	
	for(int i=0; i<(int)rules_.size(); i++) {
		short rule = rules_[i];
		//printf("In classify. Rule: %d\n", rule);	

		switch(rule) {
		case CLASSIFY_RULES_PROTOCOL:
			if( iph->ip_p != protocol_ )
				return false;
			break;
		case CLASSIFY_RULES_PORT:
		case CLASSIFY_RULES_SPORT:
		case CLASSIFY_RULES_DPORT:
			if( iph->ip_p == IPPROTO_TCP) {
				sport = ntohs(tcph->source);
				dport = ntohs(tcph->dest);
			}
			else if(iph->ip_p == IPPROTO_UDP) {
				sport = ntohs(udph->source);
				dport = ntohs(udph->dest);
			}

		//	cout << "port rule: " << rule << endl;
		//	cout << "sport: " << sport << " " << sport_ << endl; 
		//	cout << "dport: " << dport << " " << dport_ << endl;

			if(rule == CLASSIFY_RULES_PORT) {
				if(sport != port_ && dport != port_)
					return false;			
			}
			else if(rule == CLASSIFY_RULES_SPORT) {
				if(sport != sport_)
					return false;
			}
			else if(rule == CLASSIFY_RULES_DPORT) {
				if(dport != dport_)
					return false;	
			}

			break;

		case CLASSIFY_RULES_REGULAREXPRESSION:
			//cout << "Reg test" << endl;
			if(!classify_by_regex(pkt, pkt_len) )
				return false;
			//cout << "Pass reg test" << endl;
			break;		
		case CLASSIFY_RULES_TCPFLAG:	// ************ This is a trick to save time now **************
			cout << tcph->fin << " " << tcph->rst << endl;
			if(!tcph->fin && !tcph->rst)
				return false;
			break;
		case CLASSIFY_RULES_USERFUNC:
			//cout << "Call CLASSIFY_RULES_USERFUNC" << endl;
			if( func_classify == NULL )
				return false;

			return (*func_classify)(pkt, pkt_len);		
			break;
		default:
			break;
		}

		/*
		if(rule == CLASSIFY_RULES_PROTOCOL) {
			if( iph->ip_p != protocol_ )
				return false;
		}
		
		if(rule == CLASSIFY_RULES_PORT || rule == CLASSIFY_RULES_SPORT || rule == CLASSIFY_RULES_DPORT) {			
			if( iph->ip_p == IPPROTO_TCP) {
				sport = ntohs(tcph->source);
				dport = ntohs(tcph->dest);
			}
			else if(iph->ip_p == IPPROTO_UDP) {
				sport = ntohs(udph->source);
				dport = ntohs(udph->dest);
			}

		//	cout << "port rule: " << rule << endl;
		//	cout << "sport: " << sport << " " << sport_ << endl; 
		//	cout << "dport: " << dport << " " << dport_ << endl;

			if(rule == CLASSIFY_RULES_PORT) {
				if(sport != port_ && dport != port_)
					return false;			
			}
			else if(rule == CLASSIFY_RULES_SPORT) {
				if(sport != sport_)
					return false;
			}
			else if(rule == CLASSIFY_RULES_DPORT) {
				if(dport != dport_)
					return false;	
			}
				
		}

		if(rule == CLASSIFY_RULES_REGULAREXPRESSION) {
			//cout << "Reg test" << endl;
			if(!classify_by_regex(pkt, pkt_len) )
				return false;
			//cout << "Pass reg test" << endl;
		}
		**/
	}
		
	return true;
}

bool ClassifyRule::classify_by_regex(char *pkt, int pkt_len)
{
	struct ip *iph = (struct ip*) pkt;
	//struct udphdr *udph;
	//struct tcphdr *tcph;
	int ret;
	CharVect cv;
	bool bmatch = false;

	char *payload = NULL;
	int payload_len;

	if( start_pos_ == REGEX_START_PAYLOAD ) {
		payload = ip_payload(iph, pkt_len, payload_len);
	}
	else if( start_pos_ == REGEX_START_TCP) {
		payload =  ((char *)iph) + iph->ip_hl*4;
		payload_len = pkt_len - iph->ip_hl*4;
	}
	else if( start_pos_ == REGEX_START_IP) {
		// do nothing
	}	

	char *paystr = new char[ payload_len + 1 ];
	memcpy(paystr, payload, payload_len);
	paystr[payload_len] = 0;

	for(int i=0; i<payload_len; i++) {
		if(paystr[i] < ' ')
			paystr[i] = ' ';
		//cout << paystr[i] << ' ';
		//printf("%d ", paystr[i]);
	}
	//cout << endl;
			
	size_t nmatch = 1;
	regmatch_t pmatch[1];
	
	//cout << "payload: " << paystr << "......."<< endl;
	ret = regexec(&reg_, paystr, nmatch, pmatch, 0);
	//cout << pmatch[0].rm_so << " " << pmatch[1].rm_so << endl;
	//cout << pmatch[0].rm_so << " " <<  endl;
	if(ret == 0) {
		if(pmatch[0].rm_so >= 0)
			bmatch = true;					
	}	

	delete paystr;

	//cout << "Return " << bmatch << endl;
	return bmatch;
}

bool ClassifyRule::AddRegex(char *pattern, int eflag)
{
	int ret = regcomp(&reg_, pattern, REG_EXTENDED || REG_ICASE || eflag);
	if(ret)
		return false;
	return true;
}

bool ClassifyRule::ReadXML(XMLNode xSignature)
{
	XMLNode xProto, xPort, xExpr, xReg, xTCPFlag;

	xProto = xSignature.getChildNode("Protocol");
	if(!xProto.isEmpty()) {
		add_rule(CLASSIFY_RULES_PROTOCOL);
		protocol() = atoi( xProto.getText() );				
	}

	xPort = xSignature.getChildNode("Port");
	if(!xPort.isEmpty()) {
		add_rule(CLASSIFY_RULES_PORT);
		port() = atoi( xPort.getText() );				
	}

	xPort = xSignature.getChildNode("SPort");
	if(!xPort.isEmpty()) {
		add_rule(CLASSIFY_RULES_SPORT);
		sport() = atoi( xPort.getText() );				
	}

	xPort = xSignature.getChildNode("DPort");
	if(!xPort.isEmpty()) {
		add_rule(CLASSIFY_RULES_DPORT);
		dport() = atoi( xPort.getText() );				
	}

	xTCPFlag = xSignature.getChildNode("TCPFlag");
	if(!xTCPFlag.isEmpty()) {
		add_rule(CLASSIFY_RULES_TCPFLAG);
		// To be added
	}
	

	xExpr = xSignature.getChildNode("Expression");
	if(!xExpr.isEmpty()) {
		add_rule(CLASSIFY_RULES_EXPRESSION);
		//expr() = string( xExpr.getText() );				
		expr_.SetExpr(xExpr.getText());
		expr_.PrintItems();
	}

	// Regular expression
	xReg = xSignature.getChildNode("RegEx");
	if(!xReg.isEmpty()) {
		add_rule(CLASSIFY_RULES_REGULAREXPRESSION);
		//rule->classrule_.regex() = string( xReg.getText() );		
		if( ! AddRegex(const_cast<char *>(xReg.getText()), 0) ) {
			cerr << "Signature add pattern \"" << xReg.getText() << "\" error" << endl;
			return false;
		}
		
		XMLCSTR start = xReg.getAttribute("start");
		if(start) {
			string str = string(start);
			if(str == "Payload" || str == "payload" || str == "PAYLOAD") {
				start_pos() = REGEX_START_PAYLOAD;
			}
			else if(str == "tcp" || str == "Tcp" || str == "TCP" ||
					str == "udp" || str == "Ucp" || str == "UDP"
				) {
				start_pos() = REGEX_START_TCP;
			}
			else if(str == "ip" || str == "IP" || str == "Ip") {
				start_pos() = REGEX_START_IP;
			}						
		}

		//cout << "Regex: " << xReg.getText() << ", start: " << rule->classrule_.start_pos() << endl;
	}

	XMLNode xLib, xFunc;
	xLib = xSignature.getChildNode("Libray");
	xFunc = xSignature.getChildNode("Function");
	
	if( !xLib.isEmpty() && !xFunc.isEmpty()) {
		void *libc;
		if( (libc = dlopen( xLib.getText(), RTLD_LAZY))) {
			cout << "To load func " << xFunc.getText() << " from library " << xLib.getText() << endl;
			
			if( (func_classify = (Func_Classify)dlsym(libc, xFunc.getText())) == NULL ) {
				cerr << "The function " << xFunc.getText() << " cannot be found in library " << xLib.getText() << endl;
				return false;
			}

			add_rule(CLASSIFY_RULES_USERFUNC);
			//cout << "Load finished" << endl;
		}
		else {
			cerr << "The library " << xLib.getText() << " cannot be loaded: " << dlerror() << endl;
			return false;
		}
	}

	return true;
}

BasicRule::BasicRule(): type_(RULES_NONE), value_(NULL), const_(NULL), func_extract_ids(NULL)
{	
}

bool BasicRule::AddRegex(char *pattern, int eflag)
{
	int ret = regcomp(&reg_, pattern, REG_EXTENDED || REG_ICASE || eflag);
	if(ret)
		return false;
	return true;
}

CharVect BasicRule::get_id_by_regex(char *pkt, int pkt_len)
{
	struct ip *iph = (struct ip*) pkt;
	//struct udphdr *udph;
	//struct tcphdr *tcph;
	int ret;
	CharVect cv;

	char *transport_pkt = ((char *)iph) + iph->ip_hl*4, *payload = NULL;
	int payload_len = pkt_len - iph->ip_hl*4;
	
	if(iph->ip_p == IPPROTO_TCP) {
		int tcph_len = ((struct tcphdr *)transport_pkt)->doff*4;
		payload = transport_pkt + tcph_len;
		payload_len -= tcph_len;
	}
	else if(iph->ip_p == IPPROTO_UDP) {
		payload = transport_pkt + sizeof(struct udphdr);
		payload_len -= sizeof(struct udphdr);
	}

	char *paystr = new char[ payload_len + 1 ];
	memcpy(paystr, payload, payload_len);
	paystr[payload_len] = 0;
		
	//udph = (struct udphdr *) transport_pkt;
	//tcph = (struct tcphdr *) transport_pkt;


	size_t nmatch = 2;
	regmatch_t pmatch[2];
	//cout << "payload: " << paystr << "......."<< endl;
	ret = regexec(&reg_, paystr, nmatch, pmatch, 0);
	if(ret == 0) {
		//cout << nmatch << " " << pmatch[2].rm_so << " " <<pmatch[2].rm_eo <<  endl;
		for(int i = pmatch[1].rm_so; i<=pmatch[1].rm_eo; i++) {
			if( paystr[i] != '\r' && paystr[i] != '\n')	// ***********bad*************
			cv.push_back( paystr[i] );
		}			
	}	

	delete paystr;

	return cv;
}

bool BasicRule::ReadXML(XMLNode xMessage, char *rulename)
{
	XMLNode xRule = xMessage.getChildNode(rulename);
	
	if(xRule.isEmpty()) {
		cerr << "No rule named " << rulename << endl;
		return true;
	}

	rulename_ = rulename;
	return ReadXML(xRule);
}

bool BasicRule::ReadXML(XMLNode xRule)
{
	XMLNode xType, xPattern;
	char *rulename = rulename_;

	cout << "Read rule " << rulename << endl;
	
	if(xRule.isEmpty()) {
		cerr << "No rule specified " << endl;
		return false;
	}
	
	xType = xRule.getChildNode("Type");
	if(!xType.isEmpty()) {
		int type = type_ = atoi( xType.getText() );							
		cout << "type " << type << endl;

		if(type == RULES_NONE || type == RULES_COPY_LINK_ID || type == RULES_INHERIT_PLINKID) {
			// Do nothing
		}
		else if(type == RULES_REGEX) { 
			xPattern = xRule.getChildNode("Pattern");

			if(xPattern.isEmpty()) {
				cerr << "No regular expression specified in rule " << rulename << endl;
				return false;
			}
			
			char *value = stringDup( xPattern.getText() );
			cout << "Pattern: " << value << endl;

			if( ! AddRegex(value, 0) ) {
				cerr << "Add pattern \"" << value << "\" error" << endl;
				return false;
			}										
			freeXMLString(value);
		}
		else if(type == RULES_USERFUNC) {
			XMLNode xLib, xFunc;
			xLib = xRule.getChildNode("Libray");
			xFunc = xRule.getChildNode("Function");
			
			if( !xLib.isEmpty() || !xFunc.isEmpty()) {
				void *libc;
				if( (libc = dlopen( xLib.getText(), RTLD_LAZY))) {
					cout << "To load func " << xFunc.getText() << " from library " << xLib.getText() << endl;
					//void (*test) ();
					//if( (*(void **)func_extract_ids = dlsym(libc, xFunc.getText())) == NULL ) {
					if( (func_extract_ids = dlsym(libc, xFunc.getText())) == NULL ) {
						cerr << "The function " << xFunc.getText() << " cannot be found in library " << xLib.getText() << endl;
						return false;
					}

					//cout << "Load finished" << endl;
				}
				else {
					cerr << "The library " << xLib.getText() << " cannot be loaded: " << dlerror() << endl;
					return false;
				}
			}
			else {
				cerr << "No library or function specified for " << rulename << endl;
				return false;
			}
		}
		else if(type == RULES_CONST) {
			XMLNode xConst;
			xConst = xRule.getChildNode("Value");
			if( !xConst.isEmpty() ) {
				const_ = stringDup( xConst.getText() );
			}
			else {
				cerr << "No const value specified for" << rulename << endl;
				return false;
			}
		}
	}
	else {
		cerr << "No type specified in rule " << rulename << endl;
		return false;
	}

	return true;
}

LinkRule::LinkRule(): BasicRule(), bLink_(true)
{
	
}

/*
bool LinkRule::ReadXML(XMLNode xMessage, char *rulename)
{
	XMLNode xRule = xMessage.getChildNode(rulename);
	
	if(xRule.isEmpty()) {
		cerr << "No rule named " << rulename << endl;
		return false;
	}
	
	return ReadXML(xRule);
}
*/

bool LinkRule::ReadXML(XMLNode xRule)
{
	char *rulename = rulename_;
	
	if( !BasicRule::ReadXML(xRule) )
		return false;

	if(!xRule.isEmpty()) {
		XMLNode xbLink, xType, xValue;
		// Whether the Link_ID is used to link parent or not.
		// In some cases, Link_ID is not used to link parent, but provide the ID for Follow_ID or Return_ID
		xbLink = xRule.getChildNode("Link");
		if(!xbLink.isEmpty()) {
			int bLink = atoi( xbLink.getText() );	
			if(bLink == 0) {
				bLink_ = false;
			}
		}		
	}
	else {
		cerr << "No rule named " << rulename << endl;
		return false;
	}
	return true;
}

/*
bool LinkRule::AddRegex(char *pattern, int eflag)
{
	type_ = LINK_RULES_REGEX;
	int ret = regcomp(&reg_, pattern, REG_EXTENDED || REG_ICASE || eflag);
	if(ret)
		return false;
	return true;
}
*/

CharVect LinkRule::get_link_id(char *pkt, int pkt_len)
{
	//struct ip *iph = (struct ip*) pkt;
	//struct udphdr *udph;
	//struct tcphdr *tcph;
	CharVect cv;

	if(type_ == RULES_REGEX) {	// Regular expression		
		cv = get_id_by_regex(pkt, pkt_len);
		
	}
	else if(type_ == RULES_USERFUNC) {
		if( func_extract_ids == NULL )
			return cv;
		cout << "Call extern link_id func" << endl;
		func_link_id = (Func_Link_ID) func_extract_ids;

		return (*func_link_id)(pkt, pkt_len);
		
		//vector<CharVect> cvs = (*func_extract_ids)(pkt, pkt_len);
		//if(cvs.size() == 
		//return cvs[0];
	}
	else if(type_ == RULES_CONST) {
		cv = chars_to_charvect( const_ );
	}

	return cv;
}


FollowRule::FollowRule(): BasicRule()
{
}


vector<CharVect> FollowRule::get_follow_ids(char *pkt, int pkt_len)
{
	cout << "In get_follow_ids"<< endl;
	vector<CharVect> cvs;
	CharVect cv;

	if(type_ == RULES_REGEX) {	// Regular expression		
		cv = get_id_by_regex(pkt, pkt_len);
		cvs.push_back(cv);
		
	}
	else if(type_ == RULES_USERFUNC) {
		if( func_extract_ids == NULL )
			return cvs;
		func_follow_ids = (Func_Follow_IDs) func_extract_ids;

		return (*func_follow_ids)(pkt, pkt_len);		
	}

	return cvs;	
}

DiagnoseRule::DiagnoseRule(): max_proc_time_(0)
{
}

bool DiagnoseRule::ReadXML(XMLNode xMessage, char *rulename)
{
	XMLNode xRule = xMessage.getChildNode(rulename);
	if(xRule.isEmpty()) {
		cerr << "No rule named " << rulename << endl;
		return true;
	}
	
	XMLNode xMaxProcTime;
	xMaxProcTime = xRule.getChildNode("MaxProcessTime");
	if(!xMaxProcTime.isEmpty()) {
		max_proc_time_ = atof( xMaxProcTime.getText() );				
printf("MaxProcessTime: %g\n", max_proc_time_);
	}

	return true;
}


	
	

Rule::~Rule()
{
	u_int32_t i;
	for(i=0; i<followrules_.size(); i++)
		delete followrules_[i];
}

//////////////////////// Message /////////////////////////////

//Message::Message():seq(-1),sip(0),dip(0),sport(0),dport(0),id(0),state(NULL)
Message::Message():id(0),state(NULL)
{	
	tv1.tv_sec = tv1.tv_usec = 0;
	tv2.tv_sec = tv2.tv_usec = 0;
	ts1 = ts2 = 0;
}

bool Message::operator==(const Message &other) const 
{
	//cout << "In msg compare with pkt len " << pkt_len << " and id " << charvect_to_string(other.state->link_id()) << endl;
	if( id != other.id )
		return false;

	if(pkt_len != other.pkt_len)
		return false;

	for(int i=pkt_len-1; i>=0; i--)	// Compare from the last byte of the packets
		if(pkt[i] != other.pkt[i])
			return false;

	//cout << "Return true" << endl;
	return true;
}

/*
Message::Message(struct ip *iph):seq(-1),state(NULL)
{
	struct udphdr *udph;
	struct tcphdr *tcph;
	
	this->sip = iph->ip_src.s_addr;
	this->dip = iph->ip_dst.s_addr;	

	if(iph->ip_p == IPPROTO_UDP) {
		udph = (struct udphdr *) ( ((char *)iph) + iph->ip_hl*4);

		//msg.sport = ntohs(udph->source); 
		//msg.dport = ntohs(udph->dest);

		this->sport = ntohs(udph->source); 
		this->dport = ntohs(udph->dest);

		this->id = iph->ip_sum << 16 +  udph->check;
	}
	else if(iph->ip_p == IPPROTO_TCP) {
		tcph = (struct tcphdr *) ( ((char *)iph) + iph->ip_hl*4);

		//msg.sport = ntohs(tcph->source);
		//msg.dport = ntohs(tcph->dest);

		this->sport = ntohs(tcph->source);
		this->dport = ntohs(tcph->dest);
		this->id = iph->ip_sum << 16 +  tcph->check;
	}
	else {
		//msg.sport = msg.dport = 0;
		this->sport = this->dport = 0;

		this->id = iph->ip_sum;
	}
}
*/

///////////////////////////Class State///////////////////////////////////

State::State():index_(0),parent_state_(NULL),message_(NULL), sip_(0),dip_(0),sport_(0),dport_(0)
{
	type_ = -1;
}

State::~State()
{

	
}

CharVect State::get_transform_id(int index)
{
	if(index<0 || index >= (int)follow_ids_.size()) {
		return CharVect();
	}

	return follow_ids_[index];
}

vector<CharVect>& State::gen_transform_ids(char *pkt, int pkt_len)
{
	vector<CharVect>& ids = follow_ids_;
	CharVect cv;

	//cout << "in gen_transform_ids" << endl;
	for(int i=0; i<(int) rule_->followrules_.size(); i++) {
		FollowRule *fr = rule_->followrules_[i];

		if(fr->type() == RULES_COPY_LINK_ID) {
			if(link_id_.size() > 0)
				ids.push_back( link_id_ );
		}
		else if(fr->type() == RULES_REGEX) {
			cv = fr->get_id_by_regex( pkt, pkt_len );
			ids.push_back( cv );
		}
		else if( fr->type() == RULES_USERFUNC ) {
			vector<CharVect> cvs = fr->get_follow_ids(pkt, pkt_len);
			for(unsigned int nn = 0; nn<cvs.size(); nn++) {
				ids.push_back(cvs[nn]);
			}
		}
		else if( fr->type() == RULES_INHERIT_PLINKID) {	
			if(parent_state_) {
//cout << "Inherit parent state's link id: " << charvect_to_string( parent_state_->link_id() ) << endl;;				
				ids.push_back( parent_state_->link_id() );
			}
		}
	}
	
/*	
	if(type_ == 0) {
		CharVect cv = link_id_;

		cout << "Transformed ID: " << charvect_to_string(cv) << endl;
		ids.push_back(cv);
	}
*/
	return ids;
}

void State::Output(int level, double stime, double ltime)
{
	for(int i=0; i<level; i++)
		cout << "  "; 
	//struct ip *iph = (struct ip*) message_->pkt;
	//char *pkt = message_->pkt;
	//for(int i=0; i<40; i++)
	//	printf("%02x ", pkt[i]);
	//printf("\n");
	//printf("pkt: %p\n", pkt);
	
	//cout << rule_->msgname_ << " (" << (message_->ts1 - stime) << "," << index_ << ") (" << message_. << ")" << endl;
	cout << rule_->msgname_ << " (" << (message_->ts1 - stime) << ", " << (message_->ts1 - ltime) << ", " << index_ << ") (";
	HandlePcap::print_ip(sip_);
	//cout << sip_;
	cout << ", ";
	HandlePcap::print_ip(dip_);
	//cout << dip_;
	cout << ")";

	// check the processing time with the maximum expected processing time
	if(rule_->diagoserule_.max_proc_time() > 0) {
		if( rule_->diagoserule_.max_proc_time() < message_->ts1 - ltime ) 
			cout << " * ";
	}
	cout << endl;
}

void State::Output_Recursive(int level, double stime, double ltime)
{
	Output(level, stime, ltime);

	for(int i=0; i<(int)follow_states_.size(); i++) {
		follow_states_[i]->Output_Recursive(level+1, stime, message_->ts1);
	}
}

///////////////////////////Class Rake///////////////////////////////////

Rake::Rake()
{
	buff_time_ = 30000.0;		// default value is 3 sec. 
}

Rake::~Rake()
{
	u_int32_t i;
	
	for(i=0; i<rules_.size(); i++)
		delete rules_[i];
}

int32_t Rake::MessageID(struct ip *iph) 
{
	struct udphdr *udph;
	struct tcphdr *tcph;
	int32_t id = 0;

	if(iph->ip_p == IPPROTO_UDP) {
		udph = (struct udphdr *) ( ((char *)iph) + iph->ip_hl*4);
		id = (iph->ip_sum << 16) +  udph->check;
	}
	else if(iph->ip_p == IPPROTO_TCP) {
		tcph = (struct tcphdr *) ( ((char *)iph) + iph->ip_hl*4);
		id = (iph->ip_sum << 16) +  tcph->check;
	}
	return id;
}

void Rake::RemoveOldMessage(struct timeval &now)
{
	Message *ptmsg = NULL;
	State *ptstate = NULL;
	
	//cout << "*****Start remove old msg" << endl;
	while(pml_.size() > 0) {		// Remove old messages			
		//cout << "Message queue size " << pml_.size() << endl;
		ptmsg = pml_.front();
						
		if(  time_minus(now,  ptmsg->tv1) > buff_time_) {
			//ml.pop_front();
			pml_.pop_front();
			msg_set_.erase(ptmsg);

			ptstate = ptmsg->state;
			//printf("Free memory at %p\n", ptmsg);
			delete ptmsg;

			// Remove the states as well
			hash_map<CharVect, PStateVect>::iterator hmiter;
			for(int j=0; j<(int)ptstate->follow_ids_.size(); j++) {		
				//printf("To remove state with with pos %p and follow_id %s\n", ptstate, charvect_to_string(ptstate->follow_ids_[j]).c_str());
				printf("To remove state with with follow_id %s\n", charvect_to_string(ptstate->follow_ids_[j]).c_str());
				if( (hmiter = id_states_map_.find( ptstate->follow_ids_[j] )) != id_states_map_.end() ) {			
					PStateVect& sv = hmiter->second;			
					int len1 = sv.size();

					for(PStateVect::iterator psiter = sv.begin(); psiter != sv.end(); psiter++) {
						if( (*psiter) == ptstate ) {
							sv.erase( psiter );
							cout << "Find and remove follow_id "  << charvect_to_string(ptstate->follow_ids_[j]) << endl;
							break;
						}
					}
					
					cout << "Before erase, len: " << len1 << ", and after erase, len: " << sv.size() << endl;
					if( sv.size() == 0) {	// Remove the id
						id_states_map_.erase(hmiter);
						cout << "Remove id " << charvect_to_string( ptstate->follow_ids_[j] ) << endl;
					}
				}
				else {
					cerr << "Error: didn't find id " << charvect_to_string( ptstate->follow_ids_[j] ) << endl;
					exit(0);
				}
			}
			delete ptstate;
		}
		else {
			//printf("Queue size: %d\n", pml_.size());
			break;			
		}
	}
	//cout << "******Finish remove old msg" << endl;
}


// Parse the packet and get the Message and State
int Rake::ParseMessage(const struct pcap_pkthdr &hdr, char *pkt, int pkt_len, Message *pmsg, State *state)
{
	struct ip *iph = (struct ip*) pkt;
	struct udphdr *udph = (struct udphdr *) (pkt + iph->ip_hl*4);
	struct tcphdr *tcph = (struct tcphdr *) (pkt + iph->ip_hl*4);
	//char *payload = NULL;
	//int payload_len = 0;	
	//bool bValidMsg = true;
	
	// Remove packets with no payload and no special TCP flags
	if( (iph->ip_p == IPPROTO_TCP && ntohs(iph->ip_len) == iph->ip_hl*4 + tcph->doff*4 && 
		!(tcph->fin | tcph->syn | tcph->rst) ) ||
		(iph->ip_p == IPPROTO_UDP && (unsigned int)pkt_len == iph->ip_hl*4 + sizeof(struct udphdr) ) )
		return -1;

	//cout << "pkt len " << pkt_len << " tcp len " << tcph->doff << " ip len " << ntohs(iph->ip_len) << endl; 
/*
	for(int mm=0; mm <60; mm++) {
		unsigned char c = *(((char *)pkt)+mm);
		printf("%02x ", c);
		if(mm % 16 == 15)
			printf("\n");
	}
	printf("\n\n");
*/

	pmsg->pkt = pkt;
	pmsg->pkt_len = pkt_len;
	pmsg->tv1 = hdr.ts;
	pmsg->ts1 = time_to_double(hdr.ts);
	pmsg->id = MessageID(iph);
	//printf("New message hash ID: %u\n", pmsg->id);

	state->sip_ = iph->ip_src.s_addr;
	state->dip_ = iph->ip_dst.s_addr;
	if(iph->ip_p == IPPROTO_TCP) {
		state->sport_ = ntohs(tcph->source);
		state->dport_ = ntohs(tcph->dest);
	}
	else if(iph->ip_p == IPPROTO_UDP) {
		state->sport_ = ntohs(udph->source);
		state->dport_ = ntohs(udph->dest);
	}

	//printf("In ParseMessage. Protocol: %d\n", iph->ip_p );

	for(int i=0; i<(int)rules_.size(); i++) {
		Rule *rule = rules_[i];
		//cout << "Consider rule " << i << endl;
		if(rule->classrule_.classify(pkt, pkt_len) == true) {
			state->type_ = i;
			state->rule_ = rule;

			//cout << "Classify message as type " << i << endl;
			return i;
		}
	}

	return -1;
	
/*
	
	if(iph->ip_p == IPPROTO_UDP) {
		udph = (struct udphdr *) ( ((char *)iph) + iph->ip_hl*4);

	//	pmsg->sport = ntohs(udph->source); 
	//	pmsg->dport = ntohs(udph->dest);

		bValidMsg = false;
	}
	else if(iph->ip_p == IPPROTO_TCP) {
		tcph = (struct tcphdr *) ( ((char *)iph) + iph->ip_hl*4);

	//	pmsg->sport = ntohs(tcph->source);
	//	pmsg->dport = ntohs(tcph->dest);

		pmsg->id = tcph->seq + tcph->check;

		payload = ((char *)tcph) + sizeof(struct tcphdr);
		payload_len = pkt_len - sizeof(struct tcphdr) - iph->ip_hl*4; 
		
		if( pmsg->sport == 0 || pmsg->dport == 0) {	// The IRC port
			CharVect cv;
			//char c;
			int pos = 0;

			char *str = new char[payload_len+1], *msg_str;
			memcpy(str, payload, payload_len);
			str[payload_len] = '\0';

			if( (msg_str = strstr(str, "PRIVMSG")) != NULL ) {
				msg_str+= 7;
				while(*msg_str == ' ')
					msg_str++;
				
				//while(pos < payload_len && (c = payload[pos++]) != 0 ) {
				while(msg_str[pos] != 0 && msg_str[pos] != 0x0a && msg_str[pos] != 0x0d) {
					cv.push_back( msg_str[pos]);
					pos++;
				}

				cout << "ID: " << charvect_to_string(cv) << endl;				
				state->link_id_ = cv;	
			}
			else
				bValidMsg = false;						
		}		
	}
	else {
		//msg.sport = msg.dport = 0;
		pmsg->sport = pmsg->dport = 0;

		bValidMsg = false;
	}

	return bValidMsg;
*/

}

//
// Parameters:
//		pkt: starts from IP header
//
void Rake::AddMessage(const struct pcap_pkthdr &hdr, char *pkt, int pkt_len, int pindex)
{
	struct ip *iph = (struct ip*) pkt;
	//struct udphdr *udph;
	//struct tcphdr *tcph;
	struct timeval now = hdr.ts;

	State *state = new State();
	Message *pmsg = new Message();
	//printf("New message %p\n", pmsg);
	//bool bValidMsg;
	int msg_type = -1;

	// Remove old messages and add the message to the message list		
	RemoveOldMessage(now);
	
	msg_type = ParseMessage(hdr, pkt, pkt_len, pmsg, state);
	pmsg->state = state;
	state->message_ = pmsg;
	state->index_ = pindex;

	if(msg_type == -1) {	// The message is not matched with any predefined message in xml file
		//cout << "Non related message" << endl;
		delete pmsg;
		delete state;
	}
	else {		
		printf("\nNow process packet of seq %d and msg type: %s\n", pindex, state->rule_->msgname_.c_str());
		//cout << "Get message of type " << msg_type << endl;		
		PMsgList::iterator it;
		PMsgList::reverse_iterator rit;	
		//Message *ptmsg;
		//State *ptstate;
	
		// Check if the message matches with previous message (so this message is actually an incoming message and the previous one is the outgoing one)
		PMessageSet::iterator pmsg_it;
		if((pmsg_it = msg_set_.find(pmsg)) != msg_set_.end() ) {
			// The message is duplicated
			Message *orig_msg = *pmsg_it;

			// Update the time of the first copy
			orig_msg->tv2 = pmsg->tv1;
			orig_msg->ts2 = pmsg->ts1;

			cout << "Duplicate message: " << charvect_to_read_string(orig_msg->state->link_id_) << endl; 

			delete pmsg;
			delete state;
		}
		else {			
			// The message is a new message. So link the states
			hash_map<CharVect, PStateVect>::iterator iter;
			bool bLinked = false, bFollowed = false;
			
			pml_.push_back(pmsg);		// Add the new message to message queue
			msg_set_.insert(pmsg);

			// Response_ID
			if( state->rule_->responserule_.type() != RULES_NONE ) {
				//bool bLinkReturn = false;
				cout << "Response rule: " << state->rule_->responserule_.type() << endl;
				state->return_id_ = state->rule_->responserule_.get_link_id(pkt, pkt_len);

				cout << "Get response ID: \"" << charvect_to_read_string(state->return_id_) << "\" " << endl;
				if( (iter = return_id_states_map_.find( state->return_id_ )) != return_id_states_map_.end() ) {
					cout << "Find response ID is: \"" << charvect_to_read_string(state->return_id_) << "\" " <<state->return_id_.size() << endl;
			
					PStateVect& sv = iter->second;
					// Find the triggering state of the current state
					for(int i=(int)sv.size()-1; i>=0; i--) {
						if( sv[i]->dip_ == state->sip_ && sv[i]->sip_ == state->dip_ && sv[i]->sport_ == state->dport_ && sv[i]->dport_ == state->sport_) {	// Using the same socket
							sv[i]->follow_states_.push_back(state);
							cout << "Link msg by Response_ID" << endl;
							bLinked = true;
							break;
						}
					}								
				}
			}

			// Query_ID
			if( state->rule_->queryrule_.type() != RULES_NONE ) {
				PStateVect sv;
				cout << "Query rule: " << state->rule_->queryrule_.type() << endl;
				state->return_id_ = state->rule_->queryrule_.get_link_id(pkt, pkt_len);
				cout << "Get query ID: \"" << charvect_to_read_string(state->return_id_) << "\" " << endl;
				
				sv.push_back( state );
				return_id_states_map_[state->return_id_] = sv;
				bFollowed = true;
			}

			// Get the link_ID
			if(state->rule_->linkrule_.type() == RULES_NONE) {
				// Do noting
			}
			if( state->rule_->linkrule_.type() == RULES_SEARCH_IN_POOL) {
				// Have to search the link ID from all previously stored follow IDs
				int payload_len = 0;
				char *payload = ip_payload(iph, pkt_len, payload_len);
				
				hash_map<CharVect, PStateVect>::iterator iter;
				cout << "To search ID among the follow_ID pool" << endl;
				
				for( iter = id_states_map_.begin(); iter != id_states_map_.end(); iter++ ) {
					const CharVect &cvr = iter->first;
					int pp;
					if( (pp = hexfind(payload, payload_len, cvr) >= 0) ) {
						state->link_id_ = cvr;
						cout << "Hexfind at pos " << pp << endl;
						break;
					}																		
				}				
			}
			else 
				state->link_id_ = state->rule_->linkrule_.get_link_id(pkt, pkt_len);		// Get the link_id_	
			
			if(state->link_id_.size() == 0) {
				if(state->rule_->linkrule_.type() != RULES_NONE) {
					cout << "Failed to get Link_ID in the message" << endl;
				}
			}
			else {
				// Match parent message by Link_ID
				cout << "Got link_id: \"" << charvect_to_read_string(state->link_id_) << "\""<< endl;

				//cout << "To find msg ID: \"" << charvect_to_string(state->link_id_) << "\" " <<state->link_id_.size() << endl;
				if( (iter = id_states_map_.find( state->link_id_ )) != id_states_map_.end() ) {
					cout << "Find msg. ID is: \"" << charvect_to_read_string(state->link_id_) << "\" " <<state->link_id_.size() << endl;
				
					PStateVect& sv = iter->second;
					// Find the triggering state of the current state
					for(int i=(int)sv.size()-1; i>=0; i--) {
						if( sv[i]->dip_ == state->sip_ ) {	// Link these two states
							sv[i]->follow_states_.push_back(state);
							state->parent_state_ = sv[i];
							cout << "Link msg" << endl;
							bLinked = true;
							break;
						}
					}								
				}
				else {
					
				}

			}	
			
			// Specify the IDs (states and messages) that will triggered by the current message
			bFollowed |= AddTriggerIDs( state->gen_transform_ids(pkt, pkt_len), state);

			if(!bLinked && bFollowed) {
				start_states_.push_back( state );
			}
		}
	}	
}

bool Rake::AddTriggerIDs(const vector<CharVect> &triggered_ids, State *state) 
{
	bool bHasFollowID = false;
	hash_map<CharVect, PStateVect>::iterator iter;
	for(int j=0; j<(int)triggered_ids.size(); j++) {		
		cout << "Add trigger id: \"" << charvect_to_read_string( triggered_ids[j] ) << "\" " << triggered_ids[j].size() << ":  "; 
		if( (iter = id_states_map_.find( triggered_ids[j] )) != id_states_map_.end() ) {		
			PStateVect& sv = iter->second;			
			sv.push_back( state );
			//printf( "append. state pos: %p and new size is %d\n", state, sv.size());
			printf( "append. new size is %d\n",  sv.size());
			bHasFollowID = true;
		}
		else {
			PStateVect sv;
			sv.push_back( state );
			id_states_map_[ triggered_ids[j] ] = sv;
		//	id_states_map2_[charvect_to_string( triggered_ids[j] )] = sv;
			//printf( "new map. state pos: %p and sv size is %d\n", state, sv.size());
			printf( "new map. sv size is %d\n",  sv.size());
			bHasFollowID = true;
		} 
	}

	return bHasFollowID;
/*
	int ii=0;
	hash_map<string, PStateVect>::iterator iter2;
	cout << "Print map: " << id_states_map_.size() << endl;
	for(iter2 = id_states_map2_.begin(); iter2 != id_states_map2_.end(); ++iter2) {
		//cout << charvect_to_string( iter2->first ) << endl;
		cout << iter2->first << endl;
		if(ii++>5)
			break;
	}
	cout << endl;
*/	
}

/*
bool Rake::XML_Link_ID(XMLNode &xLinkID)
{
	
}
*/

void Rake::Output()
{
	for(int i=0; i<(int)start_states_.size(); i++) {
		start_states_[i]->Output_Recursive(0, start_states_[i]->message_->ts1, start_states_[i]->message_->ts1);
		cout << endl << endl;
	}
}

bool Rake::ReadXML(char *fname)
{
	int n1=0, n2=0, niter1, niter2;
	int i;

	XMLNode xMainNode=XMLNode::openFileHelper(fname,"Rake");
	XMLNode xMessage;

	n1=xMainNode.nChildNode("Message");	
	niter1 = 0;
	for (i=0; i<n1; i++) {  // For each state
		xMessage = xMainNode.getChildNode("Message",&niter1);

		// Create the rule (semantics)
		Rule *rule = new Rule();
		char *msgName = stringDup(xMessage.getAttribute("name"));
		rule->msgname_ = string(msgName);
		cout << "Msg name " << msgName << endl;

		// Classification rule (Signature)
		XMLNode xSignature = xMessage.getChildNode("Signature");
		if(!xSignature.isEmpty()) {
			if( !rule->classrule_.ReadXML(xSignature) )
				return false;
			
		}
		cout << "Signature rules number: " << rule->classrule_.rules().size() << endl;
		

		freeXMLString(msgName); 
		
		if( !rule->linkrule_.ReadXML(xMessage, "Link_ID") )
			return false;

		// Follow_ID
		n2 = xMessage.nChildNode("Follow_ID");	
		niter2 = 0;
		for (int jj=0; jj<n2; jj++) {  // For each state
			XMLNode xFollowID;
			FollowRule *fr = new FollowRule();
			fr->rulename_ = "Follow_ID";
			
			xFollowID = xMessage.getChildNode("Follow_ID", &niter2);

			if( ! fr->ReadXML(xFollowID) )
				return false;

			rule->followrules_.push_back(fr);
		}		

		//if( !rule->returnrule_.ReadXML(xMessage, "Return_ID") )
		//	return false;		
		/*
		if( !rule->queryrule_.ReadXML(xMessage, "Query_ID") )
			return false;		

		if( !rule->responserule_.ReadXML(xMessage, "Response_ID") )
			return false;		
		*/
		if( !rule->queryrule_.ReadXML(xMessage, "Query_ID"))
			return false;
		if( !rule->responserule_.ReadXML(xMessage, "Response_ID") )
			return false;		

		if( !rule->diagoserule_.ReadXML(xMessage, "Diagnose") )
			return false;
		
		rules_.push_back(rule);
	}

	return true;
}

///////////////////////////////////////////////////////////////////
/*
wap5::wap5():hp_(NULL),pcap_fname_(NULL),pcap_filter_(NULL),lamda_(1)
{
}

wap5::~wap5()
{
	if(hp_)
		delete hp_;

	for(int i=0; i<(int)hps_.size(); i++)
		delete hps_[i];
	hps_.clear();
}

bool wap5::init()
{
	if(pcap_fname_ == NULL) {
		fprintf(stderr, "Init error: no pcap file\n");
		return false;
	}

	hp_ = new HandlePcap();
	hp_->pcap_fname() = pcap_fname_;
	
	//if( hp_->open_pcap_file(pcap_fname_, pcap_filter_) ) {
	//	
	//}
	//else
	//	return false;

	return true;
}

// Obsolete
// The new function msg_linking will consider the parent probability of many files simultaneously
bool wap5::local_link_prob(HandlePcap *hp)
{
	const u_char *pkt;
	struct pcap_pkthdr hdr;
	//struct miphdr *iph;
	struct ip *iph;
	struct udphdr *udph;
	struct tcphdr *tcph;
	struct timeval tv, now;
	//Message msg, tmsg, *pmsg, *ptmsg;
	Message *pmsg, *ptmsg;
	PMsgList pml;
	//MsgList ml;
	PMsgList::iterator it;
	double time_thresh = 3;
	int seq = 0;
	
	if( hp->status() <= 0 ) {
		//fprintf(stderr, "The pcap handler hasn't been opened yet\n");
		//return false;

		if( !hp_->open_pcap_file(pcap_fname_, pcap_filter_) )
			return false;
	}

	int hdr_size = hp->pcap_hdr_size();	

	while ( (pkt = hp->next_pkt(&hdr)) ) {	
		iph = (struct ip *) (pkt + hdr_size);
		tv = hdr.ts;

	//	printf("%d.%.6d %d %d %d %d %d\n",tv.tv_sec, tv.tv_usec, iph->ip_src, iph->ip_dst, iph->ip_p);
		printf("%d.%.6d ", (int)tv.tv_sec, (int)tv.tv_usec);
		HandlePcap::print_ip(iph->ip_src);
		printf(" ");
		HandlePcap::print_ip(iph->ip_dst);
		printf(" %d ", iph->ip_p);

		//msg.sip = iph->ip_src.s_addr;
		//msg.dip = iph->ip_dst.s_addr;		
		pmsg = new Message();
		pmsg->sip = iph->ip_src.s_addr;
		pmsg->dip = iph->ip_dst.s_addr;	

		if(iph->ip_p == IPPROTO_UDP) {
			udph = (struct udphdr *) ( ((char *)iph) + iph->ip_hl*4);

			//msg.sport = ntohs(udph->source); 
			//msg.dport = ntohs(udph->dest);

			pmsg->sport = ntohs(udph->source); 
			pmsg->dport = ntohs(udph->dest);
		}
		else if(iph->ip_p == IPPROTO_TCP) {
			tcph = (struct tcphdr *) ( ((char *)iph) + iph->ip_hl*4);

			//msg.sport = ntohs(tcph->source);
			//msg.dport = ntohs(tcph->dest);

			pmsg->sport = ntohs(tcph->source);
			pmsg->dport = ntohs(tcph->dest);
		}
		else {
			//msg.sport = msg.dport = 0;
			pmsg->sport = pmsg->dport = 0;
		}

		pmsg->seq = seq++;
		//printf("%d %d\n", msg.sport, msg.dport);
		printf("%d %d %d\n", pmsg->sport, pmsg->dport, pmsg->seq);

		//msg.tv = pmsg->tv = now = tv;
		pmsg->tv = now = tv;

		//ml.push_back(msg);
		pml.push_back(pmsg);

		while(true) {
			//tmsg = ml.front();
			ptmsg = pml.front();
			
			//if(  time_minus(now,  tmsg.tv) > time_thresh ) {
			if(  time_minus(now,  ptmsg->tv) > time_thresh ) {
				//ml.pop_front();
				pml.pop_front();
				delete ptmsg;
			}
			else {
				printf("Queue size: %d\n", pml.size());
				break;
				
			}
		}

		for ( it=pml.begin() ; it != pml.end(); it++ ) {
			ptmsg = *it;

			double t = time_minus( now, ptmsg->tv );
			t = lamda_ * exp(-lamda_ * t);

			printf("Msg %d is msg %d's parent with probability %g\n", ptmsg->seq, pmsg->seq, t);
		}

		
	}

	return true;
}

void wap5::add_pcap_file(char *fname)
{
	HandlePcap *hp = new HandlePcap();
	hp->pcap_fname() = fname;

	hps_.push_back(hp);
}


bool wap5::msg_linking()
{
	HandlePcap *hp;
	const u_char *pkt;
	struct pcap_pkthdr hdr;
	//struct miphdr *iph;
	struct ip *iph;
	//struct udphdr *udph;
	//struct tcphdr *tcph;
	struct timeval tv;//, now;
	//Message msg, tmsg, *pmsg, *ptmsg;
	Message *pmsg, *ptmsg;
	PMsgList pml;
	//MsgList ml;
	PMsgList::iterator it;
	PMsgList::reverse_iterator rit;	
	int seq = 0, i;

	int num_file = hps_.size();
	PMsgList *mlists = new PMsgList[num_file];
	double time_thresh = 3, roll_interval = 1, max_stay = 6;		// should be configurable in future
	double now = 0;
		
	for(i=0; i<num_file; i++)
		if( hps_[i]->avg_casual_delay() == 0 )
			return false;

	double buf_start_time = 1e100, t, tdiff;

	PMsgList *pmls = new PMsgList[num_file];
	for(i=0; i<num_file; i++) {
		hp = hps_[i];
		pkt = hp->next_pkt(&hdr);
		iph = (struct ip *) (pkt + hp->pcap_hdr_size());
		
		t = time_to_double(hdr.ts);
		if(t < buf_start_time)
			buf_start_time = t;

		pmsg = new Message(iph);
		pmsg->ts = t;
		
		pmls[i].push_back(pmsg);
	}
	

	while(true) {
		//int num_new = 0;
		for(i=0; i<num_file; i++) {

			if( pmls[i].back()->ts - buf_start_time > max_stay )		// no need to read following packets
				continue;
			
			while ( (pkt = hps_[i]->next_pkt(&hdr)) ) {	
				iph = (struct ip *) (pkt + hp->pcap_hdr_size());
				//num_new++;

				pmsg = new Message(iph);	
				pmls[i].push_back(pmsg);
		
				t = time_to_double(hdr.ts);
				pmsg->ts = t;

				if( iph->ip_src.s_addr == hp->local_address() ) {		// outgoing packet
					// Finding and scoring parent messages	
					for(rit = pmls[i].rbegin(); rit != pmls[i].rend(); rit++) {
						ptmsg = *rit;
						tdiff = t - ptmsg->ts;
						if(tdiff > time_thresh)
							break;

						if( ptmsg->dip == hp->local_address()) {	// incoming packet
							double lamda = hp->lamda();
							double prob =  lamda*exp(-lamda_ * tdiff);
							pmsg->parents.push_back(ptmsg);
							pmsg->parent_probs.push_back(prob);

							ptmsg->children.push_back(pmsg);
							ptmsg->child_probs.push_back(prob);
						}
					}
				}
				else if(iph->ip_dst.s_addr == hp->local_address()) {		// incoming packet
					// match to the outgoing message

					
				}
				else {
					// Shouldn't be here
					fprintf(stderr, "The packet is neither incoming nor outgoing\n");
				}

				if( t - buf_start_time > max_stay )
					break;
			}
		}


		buf_start_time += roll_interval;
		// Remove the old messages in the lists (those messages with timestamp < buf_start_time)
		for(i=0; i<num_file; i++) {
			while(true) {
				pmsg = pmls[i].front();

				if(pmsg->ts < buf_start_time) {				
					pmls[i].pop_front();

					// For spontaneous message, figure out the possible paths of it
					
				}
			}
		}
		
	}

	return true;
}
*/

void usage(char *pname)
{
	printf("Usage: %s XML_config Pcap_trace  \n"
	       "Read the XML configure file and run the customized rake\n"
	       "\n"
	       "-h		list this help message\n"
		, pname);

	exit(0);
}


int main(int argc, char* argv[])
{
	char *pcap_filter = 0;
	char ch;
	char *prog = argv[0];
	Rake rake;
	
	while ((ch = getopt(argc, argv, "f:h")) != -1) {
		switch (ch) {
		case 'f':
			if (strcmp(optarg, "") == 0) {
				fprintf(stderr, "Invalid parameter followed by -%c\n", ch);
				exit(1);  
			}
			
			pcap_filter = optarg;			
			break;		
		case 'h':
			usage(argv[0]);
			break;
		}
	}

	argc -= optind;
	argv += optind;

	if(argc > 0) {
		
		HandlePcap hp;
		struct pcap_pkthdr pcap_hdr;
		char *pkt = NULL;

		if(argc == 2) {
			if( !rake.ReadXML(argv[0]) )
				return -1;
			//return 0;
		}
		else {
			usage(prog);
		}
		
		hp.pcap_fname() = argv[1];
		if( hp.open_pcap_file() ) {
			int nn  =0;
			while( (pkt = (char *)hp.next_pkt(&pcap_hdr)) ) {
				nn++;
				//if(nn < 16251)
				//	continue;
				//if(nn > 16265)
				//	break;
				pkt += hp.pcap_hdr_size();
				printf("\nNow process packet of seq %d\n", nn);
				rake.AddMessage(pcap_hdr, pkt, pcap_hdr.caplen - hp.pcap_hdr_size(), nn);
			}

			rake.Output();

			//cout << "When done, msg_set_: " << rake.msg_set_.size() << ", id_states_map_: " << rake.id_states_map_.size() << ", pml_ " << rake.pml_.size() << endl;
		}
		else {
			cerr << "Cannot open pcap file " << argv[0] << endl;
		}

/*
		HandlePcap hp;
		hp.pcap_fname() = argv[0];
		in_addr_t lip = hp.determine_local_address();
		printf("Local ip is %s\n", inet_ntoa( *(struct in_addr *)&lip ) );
		printf("average delay is %g\n", hp.avg_casual_delay());
		
		w.pcap_file(argv[0]);
		if( w.init() ) { 			
						
			w.local_link_prob();
		}		


		for(int i=0; i<argc; i++)
			w.add_pcap_file(argv[i]);
*/
		
	}
	else {
		usage(argv[0]);
	}
	
	return 0;
}

