/*
 * Rake
 * Author: 
 *	Yao Zhao
 * Goal:
 *	Header file of Rake
 * 
 */

#ifndef __RAKE_H__
#define __RAKE_H__

#include <sys/types.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/udp.h>
#include <netinet/tcp.h>
#include <pcap.h>
#include <iostream>
#include <sstream>
#include <string>
#include <map>
#include <list>
#include <vector>
#include <algorithm>
#include <ext/hash_map>
#include <ext/hash_set>

#define RAKE_LINKID_TYPE_REGEX		0


//#include "handle_pcap.h"

using namespace std;

class Rake;
class State;

typedef vector<char> CharVect;
/*
class CharVect
{
public:
	CharVect():data_(0),len_(0) { }
	~CharVect() { if(data_) delete data_; }
	int size() { return len_; }
	
	char *data_;
	int len_;
};
*/

#define OPERATION_PLUS 0
#define OPERATION_MINUS 1
#define OPERATION_MULTIPLY 2
#define OPERATION_DIVIDE 3
#define OPERATION_OR 4
#define OPERATION_AND 5
#define OPERATION_EQUAL 6		// =
#define OPERATION_LESS 7		// <
#define OPERATION_LESSEQUAL 8	// <=
#define OPERATION_GREATER	9	// >
#define OPERATION_GREATEREQUAL	10	// >=

#define EXPR_TYPE_OPERATION 0
#define EXPR_TYPE_INT 1
#define EXPR_TYPE_PACKET 2
struct ExprItem
{
	char type_;

	union{
		int ivalue_;
		char operation_;

		struct  {
			char start_;
			int offset_;
		} pvalue_;
		
	} data_;
};

class Expression
{
public:
	Expression() {};
	bool SetExpr(const char *expr);	
	int ReadNextItem(const char* &expr);

	void Print(const ExprItem &item);
	void PrintItems();
	
protected:	
	int read_int(const char *expr, int &value);
	int read_offset(const char *expr, int &value);
public:
	string expr_;
	vector<ExprItem> items_;
};


class Message
{
public:
	Message();
	//Message(struct ip *iph);
	
public:
//	int32_t seq;		// Sequence number for messages in the pcap file. Starts from 0
//	in_addr_t sip;
//	in_addr_t dip;
//	uint16_t sport;
//	uint16_t dport;

	int32_t id;		// An ID to differenciate the message to others by content. May use checksum

	// A message can appear twice. The first time is when the message is sent out and the 2nd time is when it is received.
	struct timeval tv1;
	double ts1;

	struct timeval tv2;
	double ts2;
	
	char *pkt;
	int pkt_len;

	State *state;
	
//	vector<Message *> parents;
//	vector<double> parent_probs;

//	vector<Message *> children;
//	vector<double> child_probs;

	bool operator==(const Message &other) const ;
};

//////////////////// Overload hash function //////////////////////
namespace __gnu_cxx                                                                                 
{                                                                                             
	template<> struct hash< std::string >                                                       
	{                                                                                           
		size_t operator()( const std::string& x ) const                                           
		{                                                                                         
			return hash< const char* >()( x.c_str() );                                              
		}                                                                                         
	};

	
	template<> struct hash<CharVect>                                                       
	{                                                                                           
		size_t operator()( const CharVect& x ) const                                           
		{                   
			int size = x.size();
			char *data = new char[size+1];			
			for(int i=0; i<(int)x.size(); i++)
				data[i] = x[i];
			data[size] = 0;
			size_t value = hash< const char* >()( data );   
			//cout << "hash " << endl;
			delete []data;
			return value;
		}  
	};
	
/*
	template<> struct hash<CharVect>                                                       
	{                                                                                           
		size_t operator()( const CharVect& x ) const                                           
		{                                                  
			char *data = new char[x.size()];			
			size_t value = hash< const char* >()( data );   
			//cout << "hash " << endl;
			delete data;
			return value;
		}  
	};
*/

	template<> struct hash<const Message *>
	{                                                                                           
		size_t operator()( Message* const &x ) const                                           
		{
			//cout << "called with id " << x->id << endl;
			return hash< int >()( x->id );                                              
		}  
	};
}

class HandlePcap;
typedef vector<HandlePcap *> HPVector;

struct EqPMessage
{
	bool operator()(const Message* m1, const Message* m2) const
	{
		return (*m1 == *m2);
	}
};


typedef list<Message> MsgList;
typedef list<Message *> PMsgList;
typedef vector<Message> MsgVector;
typedef vector<Message*> PMsgVector;


class State;
typedef vector<State> StateVect;
typedef vector<State *> PStateVect;

typedef hash_set<Message*, hash<const Message*>, EqPMessage> PMessageSet;

// The semantics of messages
class Rule;

// The rules that classify packets into different types
#define CLASSIFY_RULES_PROTOCOL				0
#define CLASSIFY_RULES_PORT					1
#define CLASSIFY_RULES_SPORT					2
#define CLASSIFY_RULES_DPORT					3
#define CLASSIFY_RULES_EXPRESSION				4
#define CLASSIFY_RULES_REGULAREXPRESSION		5
#define CLASSIFY_RULES_TCPFLAG					6
#define CLASSIFY_RULES_USERFUNC				7

#define REGEX_START_DEFAULT	0
#define REGEX_START_IP			1
#define REGEX_START_TCP		2	// TCP and UDP are the same
#define REGEX_START_UDP		2
#define REGEX_START_PAYLOAD	3
typedef bool (* Func_Classify)(char *, int);
class ClassifyRule
{
friend class Rule;
public:
	ClassifyRule();	
	bool classify(char *pkt, int pkt_len);		// Check whether the packet matches the signature or not

	char &protocol() { return protocol_; }
	int &port() { return port_; }
	int &sport() { return sport_; }
	int &dport() { return dport_; }
	//string &expr() { return expr_; }
	//string &regex() { return regex_; }
	short &start_pos() { return start_pos_; };
	bool AddRegex(char *pattern, int eflag);
	
	void add_rule(short rule) { rules_.push_back(rule); }
	vector<short> &rules() { return rules_; }
	bool classify_by_regex(char *pkt, int pkt_len);
	bool ReadXML(XMLNode xSignature);
protected:
	vector<short> rules_;	
	char protocol_;
	int port_, sport_, dport_;	
	short tcpflag_mask_;
	//string expr_;
	Expression expr_;
	
	//string regex_;
	regex_t reg_;	
	short start_pos_;

	Func_Classify func_classify;
};

// The rules that generate the IDs
#define RULES_NONE				-1
#define RULES_COPY_LINK_ID		0
#define RULES_USERFUNC			1
#define RULES_REGEX				2
#define RULES_CONST				3
#define RULES_INHERIT_PLINKID	4			// Inherit the link ID of parent message
#define RULES_SEARCH_IN_POOL	1000		// For Coral, in which we don't know the format of DHT packets and then use string matching to search ID

class BasicRule
{
friend class Rule;
friend class Rake;
public:
	BasicRule();
	virtual ~BasicRule() {};

	bool AddRegex(char *pattern, int eflag);	
	int &type() { return type_; }
	CharVect get_id_by_regex(char *pkt, int pkt_len);
	
	virtual bool ReadXML(XMLNode xMessage, char *rulename);	
	virtual bool ReadXML(XMLNode xRule);
protected:	
	char *rulename_;
	int type_;	
	char *value_;

	regex_t reg_;	

	char *const_;
	
	//vector<CharVect> (*func_extract_ids)(char *, int);
	//void (*func_extract_ids) ();
	void *func_extract_ids;

	vector<string> parent_msg_names_;
	vector<int> parent_msg_types_;
};

// The rules that generate the link_ID
//#define LINK_RULES_USERFUNC	1
//#define LINK_RULES_REGEX		2

typedef CharVect (* Func_Link_ID)(char *, int);

class LinkRule: public BasicRule
{
friend class Rule;
friend class Rake;
public:
	LinkRule();
	virtual ~LinkRule() {}
	
	//bool AddRegex(char *pattern, int eflag);
	CharVect get_link_id(char *pkt, int pkt_len);		// Get the link ID from this packet

	//int &type() { return type_; }
	bool &link() { return bLink_; }
	//virtual bool ReadXML(XMLNode xMessage, char *rulename);
	using BasicRule::ReadXML;
	virtual bool ReadXML(XMLNode xRule);
	
protected:
	bool bLink_;
	//int type_;	
	//char *value_;

	//regex_t reg_;

	//CharVect (*func_link_id)(char *, int);
	Func_Link_ID func_link_id;
};


// The rules that generate the triggered IDs by this message
//#define FOLLOW_RULES_NONE					-1
//#define FOLLOW_RULES_SAME_AS_LINK_ID		0

typedef vector<CharVect> (* Func_Follow_IDs)(char *, int);
class FollowRule: public BasicRule
{
friend class Rule;
friend class Rake;
public:
	FollowRule();
	
	//vector<CharVect> get_trigger_ids(char *pkt, int pkt_len);		// Get the link ID from this packet
	vector<CharVect> get_follow_ids(char *pkt, int pkt_len);
	
protected:

	//vector<CharVect> (*func_follow_ids)(char *, int);
	Func_Follow_IDs func_follow_ids;
};

class DiagnoseRule
{
friend class Rule;
friend class Rake;
public:
	DiagnoseRule();
	virtual ~DiagnoseRule() {}
	
	double max_proc_time() { return max_proc_time_; }
	//virtual bool ReadXML(XMLNode xMessage, char *rulename);
	//using BasicRule::ReadXML;
	virtual bool ReadXML(XMLNode xMessage, char *rulename);
	
protected:
	double max_proc_time_;	// maximum processing time
	
};


// Return rules
#define RETURN_RULES_NONE					-1

class Rule
{
friend class Rake;
friend class State;
public:
	Rule() {};
	~Rule();
	
protected:
	string msgname_;
	
	ClassifyRule classrule_;
	LinkRule linkrule_;
	vector<FollowRule *> followrules_;
	//LinkRule returnrule_;
	LinkRule queryrule_;
	LinkRule responserule_;	
	DiagnoseRule diagoserule_;
};

class State
{
friend class Rake;
public:
	State();
	virtual ~State();
	inline const CharVect& link_id() { return link_id_; }
	
protected:
	virtual CharVect get_transform_id(int index);	// The function that transform a state ID into the state ID of one of the triggered state
	virtual vector<CharVect>& gen_transform_ids(char *pkt, int pkt_len);

	void Output(int level, double stime, double ltime);
	void Output_Recursive(int level, double stime, double ltime);
	
protected:
	int index_;									// The index of the packet in pcap file. For debug
	int type_;									// The type of the state
	Rule *rule_;									// The rule for the state
	
	CharVect link_id_;							// The unique ID that the state is used to match its parent state (or the triggering state for this state)	
	CharVect return_id_;							// The response ID that the state is used to match the request
	vector<CharVect> follow_ids_;					// The unique IDs for this state
	PStateVect follow_states_;						// The triggered states by this state. 
	State* parent_state_;

	Message *message_;
	// IPs and Ports
	in_addr_t sip_;
	in_addr_t dip_;	
	short sport_;
	short dport_;

};

class Rake
{
public:
	Rake();
	~Rake();

	int32_t MessageID(struct ip *iph);
	void RemoveOldMessage(struct timeval &now);
	int ParseMessage(const struct pcap_pkthdr &hdr, char *pkt, int pkt_len, Message *pmsg, State *state);
	void AddMessage(const struct pcap_pkthdr &hdr, char *pkt, int pkt_len, int pindex);
	bool AddTriggerIDs(const vector<CharVect> &triggered_ids, State *state);
	bool ReadXML(char *fname);
	void Output();
protected:
	//PStateVect states_;

	double buff_time_;										// How large message queue Rake keeps. Default is a message queue of 3 seconds
	PMsgList pml_;											// The list of buffered messages (number of buffered messages depends on buff_time_)

	hash_map<CharVect, PStateVect> id_states_map_;			// The dictionary that map an ID to a vector of states that will trigger messages with this ID
	hash_map<CharVect, PStateVect> return_id_states_map_;		// The dictionary that map a return ID to a vector of states
	PStateVect start_states_;
	
	PMessageSet msg_set_;									// The dictionary of messages

	vector<Rule *> rules_;
};

/*
class wap5
{
public:
	wap5();
	~wap5();

	void pcap_file(char *fname) { pcap_fname_ = fname;  };
	void add_pcap_file(char *fname); 
	bool init();
	
	bool local_link_prob(HandlePcap *hp);
	bool local_link_prob() { return local_link_prob(hp_); };

	bool msg_linking();
protected:
	HandlePcap *hp_;
	HPVector hps_;

	char *pcap_fname_;
	char *pcap_filter_;

	double lamda_; 	// the paramter for exponential distributation
};
*/

#endif

