/*
 * Coral Packet Parser
 * author : Anup Goyal
 *
 * goal: Parse the network traces of coral web , dns and dhst traffic
 */

#include "pkt_sync.h"

int print_pkt = 0;
int print_id = 0;
int print_dns = 0;
int test = 0;
int print_paths = 0;
int dot_graph = 0;
int remove_dup = 0;
bool verbose = false;
char pcap_errbuf[PCAP_ERRBUF_SIZE];
#define HOSTNAME 165.124.182.78
#define DHTIDSIZE 8  

int hostIP = -1518553526;
int num_pkt = 0;
int scrhostIP = -1518553526;
char *second_filename = "-";
char *pkt_filename = "-";
char *dump_filename = "tmp.syn.pcap";
char *srchost = "165.124.182.74";
char *dsthost = "-";


int pcap_hdr_size(pcap_t * pkt_d)
{
	int dl = pcap_datalink(pkt_d);
	int hdr_size = -1;

	switch (dl) {
	case DLT_NULL:
		hdr_size = 4;
		break;

	case DLT_EN10MB:
		hdr_size = 14;
		break;

	case DLT_FDDI:
		hdr_size = 13 + 8;	// fddi_header + LLC
		break;

	case DLT_RAW:
		hdr_size = 0;
		break;

	default:
		fprintf(stderr, "unknown data link type 0x%x", dl);
		break;
	}
	return hdr_size; 
}


struct timeval time_difference(struct timeval &first, struct timeval &second) 
{ 
    struct timeval result; 
 
    if(first.tv_sec > second.tv_sec ||  
        (first.tv_sec == second.tv_sec &&  
        first.tv_usec > second.tv_usec)) { 
        result.tv_usec = first.tv_usec - second.tv_usec; 
        result.tv_sec = first.tv_sec - second.tv_sec; 
    } else { 
        result.tv_usec = second.tv_usec - first.tv_usec; 
        result.tv_sec = second.tv_sec - first.tv_sec; 
    } 
     
    /* For valid inputs, usec cannot be less than -999,999. */ 
    if(result.tv_usec < 0) { 
        result.tv_sec--; 
        result.tv_usec += 1000000; 
    } 
 
    return result; 
} 



void openPcapFile(char* &pkt_filename, char* &pcap_filter){

	pcap_t *pkt_d;
	pcap_dumper_t* dump_d;
	
	struct bpf_program bpf_prog;
	struct pcap_pkthdr hdr;
	const u_char *pkt;
	int hdr_size;

	
	if (!(pkt_d = pcap_open_offline(pkt_filename, pcap_errbuf))) {
		fprintf(stderr, "can't open file %s\n", pkt_filename);
		exit(-1);
	}

	if ((hdr_size = pcap_hdr_size(pkt_d)) < 0) {
		pcap_close(pkt_d);
		exit(-1);
	}

	if (pcap_filter) {
		if (pcap_compile(pkt_d, &bpf_prog, pcap_filter, 1, 0xffff0000))
			pcap_perror(pkt_d, "pcap_compile");
		else if (pcap_setfilter(pkt_d, &bpf_prog))
			pcap_perror(pkt_d, "pcap_setfilter");
	}


	if ( ! ( dump_d = pcap_dump_open(pkt_d, dump_filename) ) )
			{
				fprintf(stderr, "can't open file %s\n", dump_filename);
				pcap_close(pkt_d); 
				exit(-1);
			}
	


	struct timeval tv;

	while ((pkt = pcap_next(pkt_d, &hdr))) {
		++num_pkt;
		tv = hdr.ts;
		hdr.ts.tv_sec -= 1;
		pcap_dump((u_char *) dump_d, &hdr, pkt);

	}
	printf("TOTAL NUMBER OF PACKET WRITTEN IS %d\n",num_pkt);
	pcap_close(pkt_d);


}

void usage(char *pname)
{
	printf("Usage: %s -frwxzh \n"
	       "Parse the netwrok traces of coral and link the messages together\n"
	       "\n"
	       "-h		list this help message\n"
	       "-f [filter]	set a TCPDUMP filter to the input trace\n"
	       "-r [file]	read from the file. Default read from standard input (-)\n"
	       "-w [wfile]	write to the file prefix. The file names will be [wfile]-[SIP].\n"
	       "-H [IP_ADDRESS] set the reference IP address\n"
	       "-p              write pkt hexdump in stdout\n"
	       "-i              source ip\n"
	       "-d              destination host\n"
	       "-g				Write message linking graph in stdout\n"
	       "-c				write path in dot graph format\n"
	       "-m				Remove Duplicate flows\n"
	       "-v 				verbose output\n"
	       "-s 				Second pcap file \n"
	       "		Default wfile=file\n", pname);

	exit(0);
}

int main(int argc, char *argv[])
{


	char *pcap_filter = 0;


	
	map < connid_t, std::list < conntime_t > >conndict;
	map < connid_t, std::list < conntime_t > >::iterator connpos;
	std::list < conntime_t >::iterator tvit;

	map < unsigned int, vector < connrec_t > >scandict;
	map < unsigned int, vector < connrec_t > >::iterator scpos;
	vector < connrec_t >::iterator crit;


	char ch;
	while ((ch = getopt(argc, argv, "f:r:w:s:i:d:gmctpvh")) != -1) {
		switch (ch) {
		case 'f':
			if (strcmp(optarg, "") == 0) {
				fprintf(stderr, "Invalid parameter followed by -%c\n", ch);
				exit(1);
			}
			pcap_filter = optarg;
			break;
		case 'r':
			if (strcmp(optarg, "") == 0) {
				fprintf(stderr, "Invalid parameter followed by -%c\n", ch);
				exit(1);
			}
			pkt_filename = optarg;
			break;
		case 'w':
			if (strcmp(optarg, "") == 0) {
				fprintf(stderr, "Invalid parameter followed by -%c\n", ch);
				exit(1);
			}
			dump_filename = optarg;
			break;
		case 's':
			if (strcmp(optarg, "") == 0) {
				fprintf(stderr, "Invalid parameter followed by -%c\n", ch);
				exit(1);
			}
			second_filename = optarg;
			break;
		case 'p':
		  print_pkt = 1;
		  break;
		case 'i':
			if (strcmp(optarg, "") == 0) {
				fprintf(stderr, "Invalid parameter followed by -%c\n", ch);
				exit(1);
			}
			srchost = optarg;
		  print_id = 1;
		  break;
		case 'd':
			if (strcmp(optarg, "") == 0) {
				fprintf(stderr, "Invalid parameter followed by -%c\n", ch);
				exit(1);
			}
			dsthost = optarg;
		  print_dns = 1;
		  break;
		case 't':
		  test = 1;
		  break;
		case 'g':
			print_paths = 1;
			break;
		case 'c':
			dot_graph = 1;
			break;
		case 'm':
			remove_dup= 1;
			break;
		case 'v':
			verbose= true;
			break;
		case 'h':
			usage(argv[0]);
			break;
		}
	}

	argc -= optind;
	argv += optind;

	
	openPcapFile(pkt_filename, pcap_filter);
}



