/*
 * Toolkits of function
 * Author: 
 *	Yao Zhao
 * Goal:
 *	provide common functions
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/resource.h>

#include "util.h"


struct timeval time_difference(const struct timeval &first, const struct timeval &second) 
{ 
	struct timeval result; 

	if(first.tv_sec > second.tv_sec ||  
	  (first.tv_sec == second.tv_sec &&  
		first.tv_usec > second.tv_usec)) { 
		result.tv_usec = first.tv_usec - second.tv_usec; 
		result.tv_sec = first.tv_sec - second.tv_sec; 
	} else { 
		result.tv_usec = second.tv_usec - first.tv_usec; 
		result.tv_sec = second.tv_sec - first.tv_sec; 
	} 

	/* For valid inputs, usec cannot be less than -999,999. */ 
	if(result.tv_usec < 0) { 
		result.tv_sec--; 
		result.tv_usec += 1000000; 
	} 

	return result; 
} 

double time_minus(const struct timeval &first, const struct timeval &second) 
{ 
	double sec, usec; 

	usec = first.tv_usec - second.tv_usec; 
	sec = first.tv_sec - second.tv_sec; 
 
	usec /= 1000000;
	sec += usec;

	return sec; 
}

double time_to_double(const struct timeval &tv) 
{ 
	double t = tv.tv_sec;
	t += tv.tv_usec / 1000000.0;
	return t; 
}

string charvect_to_string(const vector<char> &v)
{
	string s = "";

	for(unsigned int i=0; i<v.size(); i++)
		s += v[i];

	return s;
}

string charvect_to_hexstring(const vector<char> &v)
{
	string s = "";
	char hex[3];
	hex[2] = 0;
	for(unsigned int i=0; i<v.size(); i++) {
		sprintf(hex,"%.2x", (unsigned char)v[i]);
		s += hex;
	}

	return s;
}

string charvect_to_read_string(const vector<char> &v)
{
	bool bRead = true;
	for(int i=0; i<v.size(); i++) {
		if( v[i] < 32 ) {
			bRead = false;
			break;
		}
	}

	if(bRead)
		return charvect_to_string(v);
	else
		return charvect_to_hexstring(v);
}

vector<char> chars_to_charvect(char *chars)
{
	vector<char> v;
	if( chars ==  NULL )
		return v;
	
	while(*chars != 0) {
		v.push_back(*chars);
		chars++;
	}

	return v;
} 

vector<char> string_to_charvect(string &s)
{
	return chars_to_charvect((char *)s.c_str());
}

vector<char> nchars_to_charvect(char *chars, int n)
{
	vector<char> v;
	while(n-- > 0) {
		v.push_back(*chars);
		chars++;
	}

	return v;
}

char *ip_payload(struct ip *iph)
{
	struct udphdr *udph;
	struct tcphdr *tcph;
	
	char *transport_pkt = ((char *)iph) + iph->ip_hl*4, *payload;

	if(iph->ip_p == IPPROTO_TCP) {
		int tcph_len = ((struct tcphdr *)transport_pkt)->doff*4;
		payload = transport_pkt + tcph_len;
	}
	else if(iph->ip_p == IPPROTO_UDP) {
		payload = transport_pkt + sizeof(struct udphdr);
	}

	return payload;
}


char *ip_payload(struct ip *iph, int iplen, int &payloadlen)
{
	//struct udphdr *udph;
	//struct tcphdr *tcph;
	
	char *transport_pkt = ((char *)iph) + iph->ip_hl*4, *payload;
	payloadlen = iplen - iph->ip_hl*4;

	if(iph->ip_p == IPPROTO_TCP) {
		int tcph_len = ((struct tcphdr *)transport_pkt)->doff*4;
		payload = transport_pkt + tcph_len;
		payloadlen -= tcph_len;
	}
	else if(iph->ip_p == IPPROTO_UDP) {
		payload = transport_pkt + sizeof(struct udphdr);
		payloadlen -= sizeof(struct udphdr);
	}
	else
		return 0;

	return payload;
}

void push_vect_vector(vector<char> &vd, const vector<char> &vs)
{
	for(int i=0; i<vs.size(); i++)
		vd.push_back(vs[i]);
}

void push_vect_string(vector<char> &v, string &s)
{
	push_vect_chars(v, s.c_str());
}

void push_vect_chars(vector<char> &v, const char *chars)
{
	while(*chars != 0) {
		v.push_back(*chars);
		chars++;
	}
}

void push_vect_nchars(vector<char> &v, char *chars, int n)
{
	for(int i=0; i<n; i++)
		v.push_back( chars[i] );
}

void push_vect_int(vector<char> &v, int n)
{
	char *pn = (char *)&n;
	for(int i=0; i<4; i++)
		v.push_back( pn[i] );
}

void push_vect_int_reverse(vector<char> &v, int n)
{
	//char *pn = (char *)&n;
	//for(int i=3; i>=0; i++)
	//	v.push_back( pn[i] );

	n = htonl(n);
	push_vect_int(v, n);
}


void push_vect_short(vector<char> &v, short n)
{
	char *pn = (char *)&n;
	for(int i=0; i<2; i++)
		v.push_back( pn[i] );
}

vector<char> sock_vect(struct ip *iph)
{
	vector<char> sock_id;	
	char *transport_pkt = ((char *)iph) + iph->ip_hl*4;

	push_vect_int(sock_id, iph->ip_src.s_addr);
	push_vect_int(sock_id, iph->ip_dst.s_addr);
	
	if(iph->ip_p == IPPROTO_TCP) {
		struct tcphdr *tcph = (struct tcphdr *)transport_pkt;
		
		sock_id.push_back('T');	
		push_vect_short(sock_id, tcph->source);
		push_vect_short(sock_id, tcph->dest);
	}
	else if(iph->ip_p == IPPROTO_UDP) {
		struct udphdr *udph = (struct udphdr *)transport_pkt;

		sock_id.push_back('U');
		push_vect_short(sock_id, udph->source);
		push_vect_short(sock_id, udph->dest);		
	}

	return sock_id;
}

vector<char> sock_vect_reverse(struct ip *iph)
{
	vector<char> sock_id;	
	char *transport_pkt = ((char *)iph) + iph->ip_hl*4;

	push_vect_int(sock_id, iph->ip_dst.s_addr);
	push_vect_int(sock_id, iph->ip_src.s_addr);

	if(iph->ip_p == IPPROTO_TCP) {
		struct tcphdr *tcph = (struct tcphdr *)transport_pkt;
		
		sock_id.push_back('T');			
		push_vect_short(sock_id, tcph->dest);
		push_vect_short(sock_id, tcph->source);
	}
	else if(iph->ip_p == IPPROTO_UDP) {
		struct udphdr *udph = (struct udphdr *)transport_pkt;

		sock_id.push_back('U');
		push_vect_short(sock_id, udph->dest);		
		push_vect_short(sock_id, udph->source);
	}

	return sock_id;
}


int hexfind(char *str, int len, const vector<char> &cv)
{
	// Now use the most stupid implementation
	int n = len - cv.size(), m = cv.size();
	for(int i=0; i<n; i++) {
		int pos = i;
		bool bMatch = true;
		for(int j =0; j<m; j++,pos++) {
			if(cv[j] != str[pos]) {
				bMatch = false;
				break;
			}
		}

		if(bMatch)
			return i;
	}
	return -1;
}

int read_int_java(char *str)
{
	int *pvalue, value;
	pvalue = (int *)str;

	value = *pvalue;
	return ntohl(value);
}

short read_short_java(char *str)
{
	short *pvalue, value;
	pvalue = (short *)str;

	value = *pvalue;
	return ntohs(value);
}

